import pandas as pd
"""This module contains the InternalGains generator file for EnergyPlus. It considers the
People, Light, Electricity or plugin Loads, and GasEquipment for Cooking objects in the model.
"""

# Define a class for the type of user living in the building
# The types of users are: DayAwayEveningHome, MostlyHomeEarlyReiser, DayAwayEveningAway, and MostlyHome
class UserType:
    def __init__(self, name, type):
        self.name = name
        self.type = type
        # The types of users are: DayAwayEveningHome, MostlyHomeEarlyReiser, DayAwayEveningAway, and MostlyHome
        
    # The following funtions define the occupancy profile, the activity level, the light use, the electric plugin use, and the gas equipment use.
    def occupancy_profile(self):
        """This function defines the occupancy profile of the user. For that, it read the 
        occupancy profile from the file allocated on eprllib.ExampleFiles.users and apply then
        a random modification based on stadistics.
        """
        # Read the example file with pandas using the self.type variable as a reference for the file name.
        occupancy_profile = pd.read_csv(f'../ExampleFiles/users/{self.type}.csv')
        # Apply a random modification based on stadistics.
        occupancy_profile = occupancy_profile.sample(frac=1).reset_index(drop=True)
        
    def activity_level(self):
        pass
    def light_use(self):
        pass
    def electric_plugin_use(self):
        pass
    def gas_equipment_use(self):
        pass
    def cooking_use(self):
        pass
            

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


"""# RUN CONVENTIONAL CONTROLS

This script execute the conventional controls in the evaluation scenario.
"""
import os
import csv
from eprllib.Env.MultiAgent.EnergyPlusEnvironment import EnergyPlusEnv_v0
from eprllib.Agents.ConventionalPolicy import ConventionalPolicy
from eprllib.Tools.ActionTransformers import ActionTransformer
import pandas as pd
import threading
from queue import Queue, Empty
from typing import Dict, Any, Optional

class rb_evaluation:
    
    def __init__(
        self,
        env_config: Dict[str, Any],
        policy_config: Dict[str, ConventionalPolicy],
        name: str,
    ) -> None:
        
        self.env_config = env_config
        self.policy_config = policy_config
        self.name = name
        self.env = EnergyPlusEnv_v0(env_config)
        self._agent_ids = self.env.get_agent_ids()
        self.terminated = {}
        self.terminated = False
        self.data_queue: Optional[Queue] = None
        self.data_processing: Optional[step_processing] = None
        self.timestep = 0
        
        self.policy = {agent: config for agent, config in self.policy_config.items()}
        
        if not os.path.exists(env_config['output']):
            os.makedirs(env_config['output'])
    
    def start_simulation(self) -> None:
        self.data_queue = Queue()
        self.data_processing = step_processing(
            self.data_queue,
            f"{self.env_config['output']}/{self.name}.csv",
        )
        self.data_processing.run()
        # se obtiene la observaión inicial del entorno para el episodio
        obs_dict, infos = self.env.reset()
        
        # Create an empty DataFrame to store the data
        obs_keys = self.env.energyplus_runner.obs_keys
        infos_keys = self.env.energyplus_runner.infos_keys
        
        data = ['agent_id']+['timestep']+obs_keys+['Action']+['Reward']+['Terminated']+['Truncated']+infos_keys
        # coloca los datos en una cola
        self.data_queue.put(data)
        
        prev_action = {agent: 0 for agent in self._agent_ids}
        while not self.terminated: # se ejecuta un paso de tiempo hasta terminar el episodio
            # se calculan las acciones convencionales de cada elemento
            actions_dict = {agent: 0 for agent in self._agent_ids}
            for agent in self._agent_ids:
                action = self.policy[agent].compute_single_action(infos[agent], prev_action[agent])
                actions_dict[agent] = action
                prev_action[agent] = action
            
            # Get the values of the variables for a timestep
            obs_dict, reward, terminated, truncated, infos = self.env.step(actions_dict)
            
            if self.env_config.get('action_transformer', False):
                action_transformer:ActionTransformer = self.env_config['action_transformer']
                action_transformer = action_transformer(self.env_config['agents_config'], self._agent_ids)
                # Transform all the actions
                dict_action = action_transformer.transform_action(dict_action)
            
            for agent in self._agent_ids:
                data = [agent, self.timestep] + list(obs_dict[agent]) + [actions_dict[agent], reward[agent], terminated["__all__"], truncated["__all__"]] + [value for value in infos[agent].values()]
                # coloca los datos en una cola
                self.data_queue.put(data)
            self.timestep += 1
            self.terminated = terminated["__all__"]

class step_processing:
    def __init__(
        self, 
        data_queue: Queue,
        output_path: str,
    ) -> None:
        
        self.data_queue = data_queue
        self.output_path = output_path
    
    def save_data(self) -> None:
        # Función que consume los datos de la cola y los agrega al DataFrame
        data_df = pd.DataFrame()
        
        while True:
            try:
                datos = self.data_queue.get(timeout=100)
                data_df = pd.concat([data_df, pd.DataFrame([datos])], ignore_index=True)
                # Guarda el DataFrame periódicamente o al final del episodio
                if len(data_df) >= 1000:
                    with open(self.output_path, 'a') as f:
                        data_df.to_csv(f, index=False, header=False)
                    data_df = pd.DataFrame()
            except (Empty):
                with open(self.output_path, 'a') as f:
                    data_df.to_csv(f, index=False, header=False)
                break
        # join the Thread back to the main thread, otherwise the program will close
        self.stop()
        
    def run(self) -> None:
        # Inicia un hilo para guardar los datos
        self.thread = threading.Thread(target=self.save_data, args=self)
        self.thread.start()
        
    def stop(self) -> None:
        # Detiene el hilo
        print("Stopping data processing.")
        self.thread.join()



"""The 'env' module contain the base environments to wrap an EnergyPlus model
into a Markov Desition  Process (MDP) compatible with RLlib.

There is a single agent environment and a multi-agent environment.
Each of them are compose from two parts: the `EnergyPlusRunner` and the `GymEnv`.
The first one is a wrapper for the EnergyPlus simulation, while the second one
is a wrapper for the RLlib interface (based on the OpenAI Gym interface).

The `EnergyPlusRunner` is responsible for setting up the EnergyPlus simulation.
It is a non-blocking process that runs in a separate thread. It is also
responsible for handling the output of the simulation and updating the state
of the simulation. It is also responsible for providing the observations and
actions to the RLlib environment.

The `GymEnv` is responsible for providing the observations and actions to the
RLlib agent. It is also responsible for handling the input from the RLlib agent
and updating the state of the simulation.

The `EnergyPlusRunner` and `GymEnv` are designed to be used in a non-blocking
way to allow the RLlib agent to interact with the simulation in real-time.
This is done by using a separate thread for the `EnergyPlusRunner` and a
separate process for the `GymEnv`. This allows the RLlib agent to interact
with the simulation in real-time while the simulation is running in the
background.
"""



"""# ENERGYPLUS RLLIB ENVIRONMENT

This script define the environment of EnergyPlus implemented in RLlib. To works 
need to define the EnergyPlus Runner.
"""
from ray.rllib.env.multi_agent_env import MultiAgentEnv
from queue import Empty, Full, Queue
from typing import Any, Dict, Optional
from eprllib.Env.MultiAgent.EnvUtils import env_value_inspection, obs_space, continuous_action_space, discrete_action_space
from eprllib.Env.MultiAgent.EnergyPlusRunner import EnergyPlusRunner
from eprllib.Tools import Rewards

class EnergyPlusEnv_v0(MultiAgentEnv):
    """The EnergyPlusEnv_v0 class represents a multi-agent environment for 
    reinforcement learning tasks related to building energy simulation using 
    EnergyPlus software. It inherits from the MultiAgentEnv class, which 
    suggests that it supports multiple agents interacting with the environment.

    The class initializes with an env_config dictionary that contains various 
    configuration settings for the environment, such as the list of agent IDs, 
    action spaces, observable variables, actuators, meters, and other 
    EnergyPlus-related settings.
    
    The reset method is responsible for setting up a new episode of the environment. 
    It increments the episode counter, initializes queues for communication between 
    the environment and EnergyPlus, and starts an instance of the EnergyPlusRunner
    class, which likely handles the EnergyPlus simulation.
    
    The step method is the core of the environment, where agents take actions, and 
    the environment updates its state accordingly. It processes the provided actions, 
    communicates with the EnergyPlus simulation through queues, retrieves 
    observations and information from the simulation, calculates rewards based on a 
    specified reward function, and determines if the episode should terminate or truncate.
    
    The close method is used to stop the EnergyPlus simulation when the environment is 
    no longer needed.
    
    The render method is currently a placeholder and does not perform any rendering 
    functionality.
    
    Overall, this class encapsulates the logic for running EnergyPlus simulations as 
    part of a multi-agent reinforcement learning environment, allowing agents to 
    interact with the building energy simulation and receive observations, rewards, 
    and termination signals based on their actions.
    """
    def __init__(
        self,
        env_config: Dict[str, Any]
        ):
        """The __init__ method in the EnergyPlusEnv_v0 class is responsible for 
        initializing the multi-agent environment for the EnergyPlus reinforcement 
        learning task. Here's a summary of what it does:
            * 1. It assigns the env_config dictionary, which contains various 
            configuration settings for the environment, such as agent IDs, action 
            spaces, observable variables, actuators, meters, and other EnergyPlus-related 
            settings.
            * 2. It sets the _agent_ids attribute as a set of agent IDs from the env_config.
            * 3. It assigns the action_space attribute from the env_config.
            * 4. It calculates the length of the observation space based on the number of 
            observable variables, meters, actuators, time variables, weather variables, 
            and other relevant information specified in the env_config. It then creates a 
            Box space for the observation_space attribute.
            * 5. It initializes the energyplus_runner, obs_queue, act_queue, and infos_queue
            attributes to None. These will be used later for communication between the 
            environment and the EnergyPlus simulation.
            * 6. It sets up variables for tracking the episode number (episode), timestep 
            (timestep), termination status (terminateds), and truncation status (truncateds).
            * 7. It creates a dictionary last_obs and last_infos to store the last observation 
            and information for each agent.
        
        Overall, the __init__ method sets up the necessary data structures and configurations 
        for the EnergyPlus multi-agent environment, preparing it for running simulations 
        and interacting with agents.
        """
        # asigning the configuration of the environment.
        self.env_config = env_config
        # inspection of info errors
        env_value_inspection(self.env_config)
        
        # define the _agent_ids property. This is neccesary in the RLlib configuration of MultiAgnetEnv.
        self._agent_ids = set([key for key in env_config['agents_config'].keys()])
        self._thermal_zone_ids = set([self.env_config['agents_config'][agent]['thermal_zone'] for agent in self._agent_ids])
        
        # asignation of environment action space.
        self.action_space = discrete_action_space()
        # asignation of the environment observation space.
        self.observation_space = obs_space(self.env_config, self._thermal_zone_ids)
        # super init of the base class (after the previos definition to avoid errors with _agent_ids argument).
        super().__init__()
        
        # EnergyPlusRunner class.
        self.energyplus_runner: Optional[EnergyPlusRunner] = None
        # queues for communication between MDP and EnergyPlus.
        self.obs_queue: Optional[Queue] = None
        self.act_queue: Optional[Queue] = None
        self.infos_queue: Optional[Queue] = None
        
        # ===CONTROLS=== #
        # variable for the registry of the episode number.
        self.episode = -1
        self.timestep = 0
        self.terminateds = False
        self.truncateds = False
        self.env_config['num_time_steps_in_hour'] = 0
        # dict to save the last observation and infos in the environment.
        self.last_obs = {agent: [] for agent in self._agent_ids}
        self.last_infos = {agent: [] for agent in self._agent_ids}

    def reset(
        self, *,
        seed: Optional[int] = None,
        options: Optional[Dict[str, Any]] = None
    ):
        # Increment the counting of episodes in 1.
        self.episode += 1
        # saving the episode in the env_config to use across functions.
        self.env_config['episode'] = self.episode
        # stablish the timestep counting in zero.
        self.timestep = 0
        # Condition of truncated episode
        if not self.truncateds:
            # Condition implemented to restart a new epsiode when simulation is completed and 
            # EnergyPlus Runner is already inicialized.
            if self.energyplus_runner is not None and self.energyplus_runner.simulation_complete:
                self.energyplus_runner.stop()
            # Define the queues for flow control between MDP and EnergyPlus threads in a max size 
            # of 1 because EnergyPlus timestep will be processed at a time.
            self.obs_queue = Queue(maxsize=1)
            self.act_queue = Queue(maxsize=1)
            self.infos_queue = Queue(maxsize=1)
            
            # episode_config_fn: Function that take the env_config as argument and upgrade the value
            # of env_config['epjson'] (str). Buid-in function allocated in tools.ep_episode_config
            episode_config_fn = self.env_config.get('episode_config_fn', None)
            if episode_config_fn != None:
                self.env_config = episode_config_fn(self.env_config)
            
            # Start EnergyPlusRunner whith the following configuration.
            self.energyplus_runner = EnergyPlusRunner(
                episode=self.episode,
                env_config=self.env_config,
                obs_queue=self.obs_queue,
                act_queue=self.act_queue,
                infos_queue=self.infos_queue,
                _agent_ids=self._agent_ids,
                _thermal_zone_ids=self._thermal_zone_ids,
            )
            # Divide the thread in two in this point.
            self.energyplus_runner.start()
            # Wait untill an observation and an infos are made, and get the values.
            self.energyplus_runner.obs_event.wait()
            self.last_obs = self.obs_queue.get()
            self.energyplus_runner.infos_event.wait()
            self.last_infos = self.infos_queue.get()
        
        # Asign the obs and infos to the environment.
        obs = self.last_obs
        infos = self.last_infos
        
        self.terminateds = False
        self.truncateds = False
            
        return obs, infos

    def step(self, action):
        # increment the timestep in 1.
        self.timestep += 1
        # ===CONTROLS=== #
        # terminated variable is used to determine the end of a episode. Is stablished as False until the
        # environment present a terminal state.
        terminated = {}
        truncated = {}
        # Truncate the simulation RunPeriod into shorter episodes defined in days. Default: None
        cut_episode_len = self.env_config.get('cut_episode_len', None)
        if not cut_episode_len == None:
            cut_episode_len_timesteps = cut_episode_len * 24*self.env_config['num_time_steps_in_hour']
            if self.timestep % cut_episode_len_timesteps == 0:
                self.truncateds = True
        else:
            self.truncateds = False
        # timeout is set to 10s to handle the time of calculation of EnergyPlus simulation.
        # timeout value can be increased if EnergyPlus timestep takes longer.
        timeout = self.env_config.get("timeout", 10)
        
        # simulation_complete is likely to happen after last env step()
        # is called, hence leading to waiting on queue for a timeout.
        if self.energyplus_runner.simulation_complete:
            # check for simulation errors.
            if self.energyplus_runner.failed():
                raise Exception('Simulation in EnergyPlus fallied.')
                
            
            # if the simulation is complete, the episode is ended.
            self.terminateds = True
            # we use the last observation as a observation for the timestep.
            obs = self.last_obs
            infos = self.last_infos

        # if the simulation is not complete, enqueue action (received by EnergyPlus through 
        # dedicated callback) and then wait to get next observation.
        else:
            try:
                # Send the action to the EnergyPlus Runner flow.
                self.act_queue.put(action,timeout=timeout)
                self.energyplus_runner.act_event.set()
                # Get the return observation and infos after the action is applied.
                self.energyplus_runner.obs_event.wait(timeout=timeout)
                obs = self.obs_queue.get(timeout=timeout)
                self.energyplus_runner.infos_event.wait(timeout=timeout)
                infos = self.infos_queue.get(timeout=timeout)
                # Upgrade last observation and infos dicts.
                self.last_obs = obs
                self.last_infos = infos
                
                self.energyplus_runner.act_event.set()
                # Get the return observation and infos after the action is applied.
                self.energyplus_runner.obs_event.wait(timeout=timeout)
                obs = self.obs_queue.get(timeout=timeout)
                self.energyplus_runner.infos_event.wait(timeout=timeout)
                infos = self.infos_queue.get(timeout=timeout)
                # Upgrade last observation and infos dicts.
                self.last_obs = obs
                self.last_infos = infos

            except (Full, Empty):
                # Set the terminated variable into True to finish the episode.
                self.terminateds = True
                # We use the last observation as a observation for the timestep.
                obs = self.last_obs
                infos = self.last_infos
        
        # Raise an exception if the episode is faulty.
        if self.energyplus_runner.failed():
            self.terminateds = True
            raise Exception('Simulation in EnergyPlus fallied.')
        
        # Calculate the reward in the timestep
        if self.env_config.get('reward_function', False):
            reward_function = self.env_config['reward_function']
        else:
            reward_function = Rewards.dalamagkidis_2007
        reward_dict = reward_function(self, infos)
        
        terminated["__all__"] = self.terminateds
        truncated["__all__"] = self.truncateds
        
        return obs, reward_dict, terminated, truncated, infos

    def close(self):
        if self.energyplus_runner is not None:
            self.energyplus_runner.stop()
    
    def render(self, mode="human"):
        pass



"""The multi-agent aproach allow to run multiples agents in the environment, but also
a single agent.
The standard configuration use a policy with fully shared parameters.

You can check the entire configuration of the multi-agent approach in the
`multiagent_config.py` file on ExampleFiles folder.
"""




from typing import Set, Dict
import pandas as pd
import numpy as np

def trial_str_creator(trial, name:str='eprllib'):
    """This method create a description for the folder where the outputs and checkpoints 
    will be save.

    Args:
        trial: A trial type of RLlib.
        name (str): Optional name for the trial. Default: eprllib

    Returns:
        str: Return a unique string for the folder of the trial.
    """
    return "{}_{}_{}".format(name, trial.trainable_name, trial.trial_id)

def len_episode(env_config:Dict) -> str:
    """This function is used to modify the RunPeriod longitude of a epJSON file.
    
    Args:
        epjson_file(str): path to the epJSON file.
        output_folder(str): path to the destination folder where the modified file will be saved.
        episode_len(int)[Optional]: longitude of the RunPeriod, or episode in the context of eprllib. Default is 7.
        init_julian_day(int): The initial julian day to determine the RunPeriod. Defaut is 0, that means a random choice.
        
    Return:
        str: path to the modified epJSON file.
    """
    epjson_file = env_config['epjson']
    output_folder = env_config['output']
    episode_len = env_config.get('episode_len',7)
    init_julian_day = env_config.get('init_julian_day', 0)
    # Open the epjson file
    with open(epjson_file) as epf:
        epjson_object = pd.read_json(epf)
    # Transform the julian day into day,month tuple
    if init_julian_day <= 0:
        init_julian_day = np.random.randint(1, 366-episode_len)
    init_day, init_month = from_julian_day(init_julian_day)
    # Calculate the final day and month
    end_julian_day = init_julian_day + episode_len
    end_day, end_month = from_julian_day(end_julian_day)
    # Change the values in the epjson file
    epjson_object['RunPeriod']['RunPeriod 1']['beging_month'] = init_month
    epjson_object['RunPeriod']['RunPeriod 1']['begin_day_of_month'] = init_day
    epjson_object['RunPeriod']['RunPeriod 1']['end_month'] = end_month
    epjson_object['RunPeriod']['RunPeriod 1']['end_day_of_month'] = end_day
    # Save the epjson file modified into the output folder
    df = pd.DataFrame(epjson_object)
    output_path = output_folder + f'/epjson_file_{init_julian_day}.epjson'
    df.to_json(output_path, orient='records')

    print(f"The epjson file with the RunPeriod modified was saved in: {output_path}.")

    return output_path

def from_julian_day(julian_day:int):
    """This funtion take a julian day and return the corresponding
    day and month for a tipical year of 365 days.
    
    Args:
        julian_day(int): Julian day to be transform
        
    Return:
        Tuple[int,int]: (day,month)
        
    Example:
    >>> from_julian_day(90)
    31,3
    """
    # Define the number of days in each month
    days_in_months = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    # Define the day variable as equal to julian day and discount it
    day = julian_day
    for month, days_in_month in enumerate(days_in_months):
        if day <= days_in_month:
            return (day, month + 1)
        day -= days_in_month
        
def variable_checking(
    epJSON_file:str,
) -> Set:
    """This function check if the epJSON file has the required variables.

    Args:
        epJSON_file(str): path to the epJSON file.

    Return:
        set: list of missing variables.
    """
    


"""This module contains the reward functions that different authors have proposed as well as some 
developed within the framework of the development of this library. It is expected over time to be 
able to establish reward functions that optimize the operation of buildings.

For greater flexibility of the library, all reward functions must have the following format:

    ```
    def reward_function_name(EnvObject, infos: dict) -> float:
        ...
        return reward
    ```
    
The arguments correspond to the following:
    EnvObject: It is the set of properties contained in `self` of the environment used.
    infos (Dict): This is the dictionary that is shared with the `reset()` and `step()` methods 
    of the Gymnasium library along with the observation and, if applicable, the `terminated` and 
    `truncated` values.

It has been preferred to use the `infos` dictionary and not the observation, since the latter is 
a numpy array and cannot be called by key values, which is prone to errors when developing the program 
and indexing a arrangement may change.

# Property `cut_reward_len_timesteps`
This property is used to define the number of timesteps that the reward is calculated for. Not all
the rewards function have this property, so it is not mandatory to define it.

# Property `beta_reward`
This property is used to define the value of the beta parameter of the reward function. Not all
the rewards function have this property, so it is not mandatory to define it.
"""
# Importing the neccesary libraries
from typing import Dict
from math import exp
import numpy as np

# Defining the reward functions
def dalamagkidis_2007(EnvObject, infos: Dict) -> Dict[str,float]:
    """El autor plantea una función de recompensa con tres térmicos ponderados. Cada uno de estos 
    términos corresponde a una penalidad por: disconfort, uso de energía, elevada consentración de CO2.
    Cada uno de los términos es normalizado. En el caso del disconfort, el valor máximo que puede tomar es
    100%. En el caso de la energía, se debe indicar un valor de referencia como 'energy_ref'; su valor por 
    default es 1. Para la normalización de la concentración de CO2 se toma el valor de referencia indicado
    por el autor en su trabajo; este se puede modificar con la variables 'co2_ref' y su valor de default
    es 870.
    
    En la implementación se permite optar por utilizar o no cada término de la penalización.

    Args:
        EnvObject: It is the set of properties contained in `self` of the environment used.
        infos (Dict): This is the dictionary that is shared with the `reset()` and `step()` methods 
        of the Gymnasium library along with the observation and, if applicable, the `terminated` and 
        `truncated` values.

    Returns:
        float: reward value.
        
    Configuration Example:
        ```
        from eprllib.tools.rewards import dalamagkidis_2007
        
        env_config = {
            # Your config here
            ...
            "ep_variables":{
                ...
                "occupancy": ("Zone People Occupant Count", "Thermal Zone: Living"),
                "ppd": ("Zone Thermal Comfort Fanger Model PPD", "Living Occupancy"),
            },
            "ep_meters": {
                ...
                "heating": "Heating:DistrictHeatingWater",
                "cooling": "Cooling:DistrictCooling",
            },
            "ep_actuators": {
                ...
            },
            "infos_variables": [
                ...
                "ppd", 
                'heating', 
                'cooling',
                'occupancy'
            ],
            ...
            # Reward config
            'reward_function': dalamagkidis_2007,
            'reward_function_config': {
                # cut_reward_len_timesteps: Este parámetro permite que el agente no reciba una recompensa en cada paso de tiempo, en cambio las variables para el cálculo de la recompensa son almacenadas en una lista para luego utilizar una recompensa promedio cuando se alcanza la cantidad de pasos de tiempo indicados por 'cut_reward_len_timesteps'.
                'cut_reward_len_timesteps': 1,
                # Parámetros para la exclusión de términos de la recompensa
                'comfort_reward': True,
                'energy_reward': True,
                'co2_reward': True,                
                # w1: Parámetros de ponderación para el confort.
                'w1': 0.80,
                # w2: Parámetros de ponderación para el uso de energía.
                'w2': 0.01,
                # w3: Parámetros de ponderación para la concentración de CO2.
                'w3': 0.20,                 
                # energy_ref: El valor de referencia depende del entorno. Este puede corresponder a la energía máxima que puede demandar el entorno en un paso de tiempo, un valor de energía promedio u otro.
                'energy_ref': 6805274,
                # co2_ref: Este parámtero indica el valor de referencia de consentración de CO2 que se espera tener en un ambiente con una calidad de aire óptima.
                'co2_ref': 870,
                # Nombres de las variables utilizadas en su configuración del entorno.
                'occupancy_name': 'occupancy',
                'ppd_name': 'ppd',
                'T_interior_name': 'Ti',
                'cooling_name': 'cooling',
                'heating_name': 'heating',
                'co2_name': 'co2'
            }
        }
        ```
    """
    reward_dict = {key: 0. for key in EnvObject._thermal_zone_ids}
    if not EnvObject.env_config.get('reward_function_config', False):
        ValueError('The reward function configuration is not defined. The default function will be use.')
    # thermal zone variables dicts
    cut_reward_len_timesteps_tz = {thermal_zone: 0 for thermal_zone in EnvObject._thermal_zone_ids}
    w1_tz = {thermal_zone: 0 for thermal_zone in EnvObject._thermal_zone_ids}
    w2_tz = {thermal_zone: 0 for thermal_zone in EnvObject._thermal_zone_ids}
    w3_tz = {thermal_zone: 0 for thermal_zone in EnvObject._thermal_zone_ids}
    ppd_name_tz = {thermal_zone: '' for thermal_zone in EnvObject._thermal_zone_ids}
    T_interior_name_tz = {thermal_zone: '' for thermal_zone in EnvObject._thermal_zone_ids}
    occupancy_name_tz = {thermal_zone: '' for thermal_zone in EnvObject._thermal_zone_ids}
    energy_ref_tz = {thermal_zone: 0 for thermal_zone in EnvObject._thermal_zone_ids}
    cooling_name_tz = {thermal_zone: '' for thermal_zone in EnvObject._thermal_zone_ids}
    heating_name_tz = {thermal_zone: '' for thermal_zone in EnvObject._thermal_zone_ids}
    co2_ref_tz = {thermal_zone: 0 for thermal_zone in EnvObject._thermal_zone_ids}
    co2_name_tz = {thermal_zone: '' for thermal_zone in EnvObject._thermal_zone_ids}
    
    for ThermalZone in EnvObject._thermal_zone_ids:
        # define which rewards will be considered
        comfort_reward = EnvObject.env_config['reward_function_config'][ThermalZone].get('comfort_reward', True)
        energy_reward = EnvObject.env_config['reward_function_config'][ThermalZone].get('energy_reward', True)
        co2_reward = EnvObject.env_config['reward_function_config'][ThermalZone].get('co2_reward', True)
        
        # if the EnvObject don't have the list to append the values here obtained, one list is created as a property of the EnvObject
        if not hasattr(EnvObject, 'ppd_dict') and comfort_reward:
            EnvObject.ppd_dict = {key: [] for key in EnvObject._agent_ids}
        if not hasattr(EnvObject, 'energy_dict') and energy_reward:
            EnvObject.energy_dict = {key: [] for key in EnvObject._agent_ids}
        if not hasattr(EnvObject, 'co2_dict') and co2_reward:
            EnvObject.co2_dict = {key: [] for key in EnvObject._agent_ids}
        
        # define the number of timesteps per episode
        cut_reward_len_timesteps_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('cut_reward_len_timesteps', 1)
        
        # define the ponderation parameters for thermal zone
        w1_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('w1', 0.80)
        w2_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('w2', 0.01)
        w3_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('w3', 0.20)
        
        if comfort_reward:
            ppd_name_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('ppd_name', False)
            T_interior_name_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('T_interior_name', False)
            occupancy_name_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('occupancy_name', False)
            if not ppd_name_tz[ThermalZone] or not occupancy_name_tz[ThermalZone] or not T_interior_name_tz[ThermalZone]:
                ValueError('The names of the variables are not defined')
        
        if energy_reward:
            energy_ref_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('energy_ref',False)
            cooling_name_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('cooling_name', False)
            heating_name_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('heating_name', False)
            if not energy_ref_tz[ThermalZone] or not cooling_name_tz[ThermalZone] or not heating_name_tz[ThermalZone]:
                ValueError('The names of the variables are not defined')
        
        if co2_reward:
            co2_ref_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('co2_ref',False)
            co2_name_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('co2_name', False)
            occupancy_name_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('occupancy_name', False)
            if not co2_ref_tz[ThermalZone] or not co2_name_tz[ThermalZone] or not occupancy_name_tz[ThermalZone]:
                ValueError('The names of the variables are not defined')
    
    # agent variables dicts
    cut_reward_len_timesteps_a = {agent: 0 for agent in EnvObject._agent_ids}
    w1_a = {agent: 0 for agent in EnvObject._agent_ids}
    w2_a = {agent: 0 for thermal_zone in EnvObject._agent_ids}
    w3_a = {agent: 0 for agent in EnvObject._agent_ids}
    ppd_name_a = {agent: '' for agent in EnvObject._agent_ids}
    T_interior_name_a = {agent: '' for agent in EnvObject._agent_ids}
    occupancy_name_a = {agent: '' for agent in EnvObject._agent_ids}
    energy_ref_a = {agent: 0 for agent in EnvObject._agent_ids}
    cooling_name_a = {agent: '' for agent in EnvObject._agent_ids}
    heating_name_a = {agent: '' for agent in EnvObject._agent_ids}
    co2_ref_a = {agent: 0 for agent in EnvObject._agent_ids}
    co2_name_a = {agent: '' for agent in EnvObject._agent_ids}
    
    # asign the ThermalZone values above defined to each agent
    for agent in EnvObject._agent_ids:
        agent_thermal_zone = EnvObject.env_config['agents_config'][agent]['thermal_zone']
        
        comfort_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('comfort_reward', True)
        energy_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('energy_reward', True)
        co2_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('co2_reward', True)
        
        cut_reward_len_timesteps_a[agent] = cut_reward_len_timesteps_tz[agent_thermal_zone]
        w1_a[agent] = w1_tz[agent_thermal_zone]
        w2_a[agent] = w2_tz[agent_thermal_zone]
        w3_a[agent] = w3_tz[agent_thermal_zone]
        
        if comfort_reward:
            ppd_name_a[agent] = ppd_name_tz[ThermalZone]
            T_interior_name_a[agent] = T_interior_name_tz[ThermalZone]
            occupancy_name_a[agent] = occupancy_name_tz[ThermalZone]
            
        if energy_reward:
            energy_ref_a[agent] = energy_ref_tz[ThermalZone]
            cooling_name_a[agent] = cooling_name_tz[ThermalZone]
            heating_name_a[agent] = heating_name_tz[ThermalZone]
        
        if co2_reward:
            co2_ref_a[agent] = co2_ref_tz[ThermalZone]
            co2_name_a[agent] = co2_name_tz[ThermalZone]
            occupancy_name_a[agent] = occupancy_name_tz[ThermalZone]
    
    # get the values of the energy, PPD, and CO2 from the infos dict
    for agent in EnvObject._agent_ids:
        agent_thermal_zone = EnvObject.env_config['agents_config'][agent]['thermal_zone']
        
        comfort_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('comfort_reward', True)
        energy_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('energy_reward', True)
        co2_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('co2_reward', True)
        
        if comfort_reward:
            _ppd_name = ppd_name_a[agent]
            _occupancy_name = occupancy_name_a[agent]
            ppd = infos[agent][_ppd_name]
            occupancy = infos[agent][_occupancy_name]
            if occupancy == 0:
                ppd = 0
            EnvObject.ppd_dict[agent].append(ppd)
        if energy_reward:
            _cooling_name = cooling_name_a[agent]
            _heating_name = heating_name_a[agent]
            cooling_meter = infos[agent][_cooling_name]
            heating_meter = infos[agent][_heating_name]
            EnvObject.energy_dict[agent].append(cooling_meter+heating_meter)
        if co2_reward:
            co2 = infos[agent][co2_name]
            occupancy = infos[agent][_occupancy_name]
            if occupancy == 0:
                co2 = 0
            EnvObject.co2_dict[agent].append(co2)
    
    # calculate the reward if the timestep is divisible by the cut_reward_len_timesteps.
    # if don't return 0.
    for agent in EnvObject._agent_ids:
        if EnvObject.timestep % cut_reward_len_timesteps_a[agent] == 0:
            
            agent_thermal_zone = EnvObject.env_config['agents_config'][agent]['thermal_zone']
            comfort_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('comfort_reward', True)
            energy_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('energy_reward', True)
            co2_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('co2_reward', True)
            
            if comfort_reward:
                _T_interior_name = T_interior_name_a[agent]
                _T_interior = infos[agent][_T_interior_name]
                rew1 = -w1_a[agent]*(sum(EnvObject.ppd_dict[agent])/cut_reward_len_timesteps_a[agent]/100)
                # If there are not people, only the reward is calculated when the environment is far away
                # from the comfort temperature ranges. This limits are recommended in EnergyPlus documentation:
                # InputOutput Reference p.522
                if _T_interior > 29.4:
                    rew1 += -10
                elif _T_interior < 16.7:
                    rew1 += -10
            else:
                rew1 = 0
            if energy_reward:
                rew2 = -w2_a[agent]*(sum(EnvObject.energy_dict[agent])/cut_reward_len_timesteps_a[agent]/energy_ref_a[agent])
            else:
                rew2 = 0
            if co2_reward:
                rew3 = 0
                for co2 in range(len([key for key in EnvObject.co2_dict[agent].keys()])):
                    rew3 += -w3_a[agent]*(1/(1+exp(-0.06*(EnvObject.co2_dict[agent][co2]-co2_ref_a[agent]))))
                rew3 = rew3/cut_reward_len_timesteps_a[agent]
            else:
                rew3 = 0
            
            reward_dict[agent] = rew1 + rew2 + rew3
            
            # emptly the lists
            if comfort_reward:
                EnvObject.ppd_dict = {key: [] for key in EnvObject._agent_ids}
            if energy_reward:
                EnvObject.energy_dict = {key: [] for key in EnvObject._agent_ids}
            if co2_reward:
                EnvObject.co2_dict = {key: [] for key in EnvObject._agent_ids}
            
    return reward_dict
    
def normalize_reward_function(EnvObject, infos: Dict) -> Dict[str,float]:
    """This function returns the normalize reward calcualted as the sum of the penalty of the energy 
    amount of one week divide per the maximun reference energy demand and the average PPD comfort metric
    divide per the maximal PPF value that can be take (100). Also, each term is divide per the longitude
    of the episode and multiply for a ponderation factor of beta for the energy and (1-beta) for the comfort.
    Both terms are negatives, representing a penalti for demand energy and for generate discomfort.

    Args:
        self (Environment): RLlib environment.
        obs (dict): Zone Mean Air Temperature for the Thermal Zone in °C.
        infos (dict): infos dict must to provide the occupancy level and the Zone Mean Temperature.

    Returns:
        float: reward normalize value
    """
    reward_dict = {key: 0. for key in EnvObject._agent_ids}
    if not EnvObject.env_config.get('reward_function_config', False):
        ValueError('The reward function configuration is not defined. The default function will be use.')
    
    # if the EnvObject don't have the list to append the values here obtained, one list is created as a property of the EnvObject
    if not hasattr(EnvObject, 'ppd_dict'):
        EnvObject.ppd_dict = {key: [] for key in EnvObject._agent_ids}
    if not hasattr(EnvObject, 'energy_dict'):
        EnvObject.energy_dict = {key: [] for key in EnvObject._agent_ids}
    
    # variables dicts
    beta_reward_tz = {thermal_zone: 0 for thermal_zone in EnvObject._thermal_zone_ids}
    cut_reward_len_timesteps_tz = {thermal_zone: 0 for thermal_zone in EnvObject._thermal_zone_ids}
    ppd_name_tz = {thermal_zone: '' for thermal_zone in EnvObject._thermal_zone_ids}
    T_interior_name_tz = {thermal_zone: '' for thermal_zone in EnvObject._thermal_zone_ids}
    occupancy_name_tz = {thermal_zone: '' for thermal_zone in EnvObject._thermal_zone_ids}
    cooling_energy_ref_tz = {thermal_zone: 0 for thermal_zone in EnvObject._thermal_zone_ids}
    heating_energy_ref_tz = {thermal_zone: 0 for thermal_zone in EnvObject._thermal_zone_ids}
    cooling_name_tz = {thermal_zone: '' for thermal_zone in EnvObject._thermal_zone_ids}
    heating_name_tz = {thermal_zone: '' for thermal_zone in EnvObject._thermal_zone_ids}
    
    for ThermalZone in EnvObject._thermal_zone_ids:
        # define which rewards will be considered
        comfort_reward = EnvObject.env_config['reward_function_config'][ThermalZone].get('comfort_reward', True)
        energy_reward = EnvObject.env_config['reward_function_config'][ThermalZone].get('energy_reward', True)
        
        # define the number of timesteps per episode
        cut_reward_len_timesteps_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('cut_reward_len_timesteps', 1)
        
        # define the beta reward
        beta_reward_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('beta_reward', 0.5)
    
        if comfort_reward:
            ppd_name_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('ppd_name', False)
            T_interior_name_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('T_interior_name', False)
            occupancy_name_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('occupancy_name', False)
            if not ppd_name_tz[ThermalZone] or not occupancy_name_tz[ThermalZone] or not T_interior_name_tz[ThermalZone]:
                ValueError('The names of the variables are not defined')
        
        if energy_reward:
            cooling_energy_ref_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('cooling_energy_ref', False)
            heating_energy_ref_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('heating_energy_ref', False)
            cooling_name_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('cooling_name', False)
            heating_name_tz[ThermalZone] = EnvObject.env_config['reward_function_config'][ThermalZone].get('heating_name', False)
            if not cooling_energy_ref_tz[ThermalZone] or not heating_energy_ref_tz[ThermalZone] or not cooling_name_tz[ThermalZone] or not heating_name_tz[ThermalZone]:
                ValueError('The names of the variables are not defined')
    
    
    beta_reward_a = {agent: 0 for agent in EnvObject._agent_ids}
    cut_reward_len_timesteps_a = {agent: 0 for agent in EnvObject._agent_ids}
    ppd_name_a = {agent: '' for agent in EnvObject._agent_ids}
    T_interior_name_a = {agent: '' for agent in EnvObject._agent_ids}
    occupancy_name_a = {agent: '' for agent in EnvObject._agent_ids}
    cooling_energy_ref_a = {agent: 0 for agent in EnvObject._agent_ids}
    heating_energy_ref_a = {agent: 0 for agent in EnvObject._agent_ids}
    cooling_name_a = {agent: '' for agent in EnvObject._agent_ids}
    heating_name_a = {agent: '' for agent in EnvObject._agent_ids}
    
    # asign the ThermalZone values above defined to each agent
    for agent in EnvObject._agent_ids:
        agent_thermal_zone = EnvObject.env_config['agents_config'][agent]['thermal_zone']
        comfort_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('comfort_reward', True)
        energy_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('energy_reward', True)
        
        cut_reward_len_timesteps_a[agent] = cut_reward_len_timesteps_tz[agent_thermal_zone]
        beta_reward_a[agent] = beta_reward_tz[agent_thermal_zone]
        
        if comfort_reward:
            ppd_name_a[agent] = ppd_name_tz[agent_thermal_zone]
            T_interior_name_a[agent] = T_interior_name_tz[agent_thermal_zone]
            occupancy_name_a[agent] = occupancy_name_tz[agent_thermal_zone]
            
        if energy_reward:
            cooling_energy_ref_a[agent] = cooling_energy_ref_tz[agent_thermal_zone]
            heating_energy_ref_a[agent] = heating_energy_ref_tz[agent_thermal_zone]
            cooling_name_a[agent] = cooling_name_tz[agent_thermal_zone]
            heating_name_a[agent] = heating_name_tz[agent_thermal_zone]
        
    # get the values of the energy, PPD, and CO2 from the infos dict
    for agent in EnvObject._agent_ids:
        agent_thermal_zone = EnvObject.env_config['agents_config'][agent]['thermal_zone']
        comfort_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('comfort_reward', True)
        energy_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('energy_reward', True)
        
        if comfort_reward:
            _ppd_name = ppd_name_a[agent]
            _occupancy_name = occupancy_name_a[agent]
            _T_interior_name = T_interior_name_a[agent]
            ppd = infos[agent][_ppd_name]
            occupancy = infos[agent][_occupancy_name]
            T_interior = infos[agent][_T_interior_name]
            if occupancy > 0:
                if infos[agent][_ppd_name] < 5:
                    ppd = infos[agent][_ppd_name] = 5
                else:
                    ppd = infos[agent][_ppd_name]
            else:
                ppd = infos[agent][_ppd_name] = 0
            # If there are not people, only the reward is calculated when the environment is far away
            # from the comfort temperature ranges. This limits are recommended in EnergyPlus documentation:
            # InputOutput Reference p.522
            if T_interior > 29.4:
                ppd = infos[agent][_ppd_name] = 100
            elif T_interior < 16.7:
                ppd = infos[agent][_ppd_name] = 100
                
            EnvObject.ppd_dict[agent].append(ppd)
        
        if energy_reward:
            _cooling_name = cooling_name_a[agent]
            _heating_name = heating_name_a[agent]
            cooling_meter = infos[agent][_cooling_name]
            heating_meter = infos[agent][_heating_name]
            
            EnvObject.energy_dict[agent].append(cooling_meter/cooling_energy_ref_a[agent]+heating_meter/heating_energy_ref_a[agent])
    
    # calculate the reward if the timestep is divisible by the cut_reward_len_timesteps.
    # if don't return 0.
    for agent in EnvObject._agent_ids:
        if EnvObject.timestep % cut_reward_len_timesteps_a[agent] == 0:
            agent_thermal_zone = EnvObject.env_config['agents_config'][agent]['thermal_zone']
            comfort_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('comfort_reward', True)
            energy_reward = EnvObject.env_config['reward_function_config'][agent_thermal_zone].get('energy_reward', True)
            
            if comfort_reward:
                ppd_avg = sum(EnvObject.ppd_dict[agent])/len(EnvObject.ppd_dict[agent])
                rew1 = -(1-beta_reward_a[agent])*(1/(1+np.exp(-0.1*(ppd_avg-45))))
            else:
                rew1 = 0
            
            if energy_reward:
                rew2 = -beta_reward_a[agent]*(sum(EnvObject.energy_dict[agent])/len(EnvObject.energy_dict[agent]))
            else:
                rew2 = 0
            
            reward_dict[agent] = rew1 + rew2
            
    # emptly the lists
    for agent in EnvObject._agent_ids:
        if EnvObject.timestep % cut_reward_len_timesteps_a[agent] == 0:
            EnvObject.ppd_dict = {key: [] for key in EnvObject._agent_ids}
            EnvObject.energy_dict = {key: [] for key in EnvObject._agent_ids}
            
    return reward_dict



"""This module contain the utils used in the Multi-agent definition files.
"""

from typing import Tuple, Dict, Any, List, Set, Optional
from gymnasium.spaces import Box, Discrete
import numpy as np

def EP_API_add_path(version:Optional[str]="23-2-0", path:Optional[str]=None):
    """This method add the EnergyPlus Python API to the system path. This allow to use the 
    EnergyPlus program in Python. The minimal version of EnergyPlus is 9.3.0 and the default
    version (and the stable one) for eprllib is 23-2-0.

    Args:
        version (str, optional): Numeric version of EnergyPlus. Defaults to "23-2-0".
        path (Optional[str], optional): Complete path to the EnergyPlus installation directory 
        if this is different that the default installation. Defaults to None.

    Returns:
        None: Print the EnergyPlus API path added to the system path.
    """
    import sys
    
    if path is not None:
        sys.path.insert(0, f"{path}")
    else:  
        os_platform = sys.platform
        if os_platform == "linux":    
            sys.path.insert(0, f"/usr/local/EnergyPlus-{version}")
        else:
            sys.path.insert(0, f"C:/EnergyPlusV{version}")
    
    return print(f"EnergyPlus API path added: {sys.path[0]}")

def env_value_inspection(env_config:Dict):
    
    """Examine that all the neccesary arguments to use in the MDP env definition exist.

    Args:
        env_config (Dict): env_config main dict.
    """
    if not env_config.get('ep_actuators', False):
        ValueError("No actuators defined in the environment configuration. Set the dictionary in"\
            "env_config['ep_actuators'] with the corresponded actuators.")
    if not env_config.get('action_space', False):
            ValueError("No action space defined in the environment configuration. Set the dictionary in"\
                "env_config['action_space'] with the corresponded action space.")

def runner_value_inspection(env_config:Dict):
    if isinstance(env_config, str):
        ValueError("env_config is a string.")
    else:
        if not env_config.get('ep_actuators', False):
            ValueError("No actuators defined in the environment configuration. Set the dictionary in"\
                "env_config['ep_actuators'] with the corresponded actuators.")
    
        if env_config.get('use_building_properties', True):
                # a loop control the existency of the building_properties
                for key in env_config['episode_config'].keys():
                    if key in [
                        'building_area', 'aspect_ratio', 'window_area_relation_north', 
                        'window_area_relation_east', 'window_area_relation_south', 
                        'window_area_relation_west', 'inercial_mass', 
                        'construction_u_factor', 'E_cool_ref', 'E_heat_ref',
                    ]:
                        pass
                    else:
                        ValueError(f'Building property {key} not found.')
    
def actuators_to_agents(agent_config:Dict[str, List]) -> Tuple[List,List,Dict,Dict]:
    """Take the ep_actuator dict and transform it to the agent, thermal zone, and actuator type dict.

    Args:
        agent_config (Dict): ep_actuator dict in the env_config.

    Returns:
        Tuple[Dict,Dict,Dict,Dict]: agent, thermal zone, and actuator type.
    """
    # get the keys and values from the ep_actuator definitions
    keys = agent_ids = list(agent_config.keys())
    values = list(agent_config.values())
    
    # Verify that all the agents have the same quantity of attributes
    for i in range(len(values)):
        assert len(values[0])==len(values[i])
    
    # compose the actuators for EnergyPlus handle
    actuator = []
    for i in range(len(keys)):
        actuator_tuple = (values[i][0], values[i][1], values[i][2])
        actuator.append(actuator_tuple)
    # asign the actuator handle config to the agent
    agents_actuators = {}
    for i in range(len(keys)):
        agents_actuators[keys[i]] = actuator[i]

    # identify the thermal zones where there are agents
    agent_thermal_zone_names = []
    for i in range(len(keys)):
        agent_thermal_zone_names.append(values[i][3])
    # asign the thermal zone to each agent
    agents_thermal_zones = {}
    for i in range(len(keys)):
        agents_thermal_zones[keys[i]] = agent_thermal_zone_names[i]
    # a list with the thermal zones with agents (without repeat)
    thermal_zone_ids = []
    for zone in agent_thermal_zone_names:
        if not zone in  thermal_zone_ids:
            thermal_zone_ids.append(zone)
    
    # agent type
    typ = []
    for i in range(len(keys)):
        typ.append(values[i][4])
    agents_types = {}
    for i in range(len(keys)):
        agents_types[keys[i]] = typ[i]
    
    agents_str = ", ".join(agent_ids)
    print(f"The environment is defined with {len(agent_ids)} agents: {agents_str}")
    
    return agent_ids, thermal_zone_ids, agents_actuators, agents_thermal_zones, agents_types

def obs_space(env_config:Dict, _thermal_none_ids:Set):
    """This method construct the observation space of the environment.

    Args:
        env_config (Dict): The environment configuration dictionary.

    Returns:
        space.Box: The observation space of the environment.
    """
    
    obs_space_len = 0
    
    # actuator state.
    if env_config.get('use_actuator_state', True):
        obs_space_len += 1
        
    # agent_indicator.
    if env_config.get('use_agent_indicator', True):
        obs_space_len += 1
        
    # agent type.
    if env_config.get('use_agent_type', True):
        obs_space_len += 1
        
    # building properties.
    if env_config.get('use_building_properties', True):
        for thermal_zone in _thermal_none_ids:
            thermal_zone_name = thermal_zone
            break
        obs_space_len += len([key for key in env_config['building_properties'][thermal_zone_name].keys()])
        
    # weather prediction.
    if env_config.get('use_one_day_weather_prediction', True):
        obs_space_len += 24*6
        
    # variables and meters.
    if env_config.get('ep_environment_variables', False):
        obs_space_len += len(env_config['ep_environment_variables'])
    
    if env_config.get('ep_thermal_zones_variables', False):
        obs_space_len += len(env_config['ep_thermal_zones_variables'])
    
    if env_config.get('ep_object_variables', False):
        for thermal_zone in _thermal_none_ids:
            thermal_zone_name = thermal_zone
            break
        obs_space_len += len([key for key in env_config['ep_object_variables'][thermal_zone_name].keys()])
        
    if env_config.get('ep_meters', False):
        obs_space_len += len(env_config['ep_meters'])
        
    if env_config.get('time_variables', False):
        obs_space_len += len(env_config['time_variables'])
        
    if env_config.get('weather_variables', False):
        obs_space_len += len(env_config['weather_variables'])
        
    # discount the not observable variables.
    if env_config.get('no_observable_variables', False):
        for thermal_zone in _thermal_none_ids:
            thermal_zone_name = thermal_zone
            break
        obs_space_len -= len(env_config['no_observable_variables'][thermal_zone_name])
        
    # construct the observation space.
    return Box(float("-inf"), float("inf"), (obs_space_len,))

def continuous_action_space():
    """This method construct the action space of the environment.
    
    Returns:
        gym.Box: Continuous action space with limits between [0,1].
    """
    return Box(low=0.0, high=1.0, shape=(1,), dtype=np.float32)

def discrete_action_space():
    """This method construct the action space of the environment.
    
    Returns:
        gym.Discrete: Discrete action space with limits between [0,10] and a step of 1.
    """
    return Discrete(11)

def environment_variables(env_config: Dict[str, Any]) -> Tuple[Dict[str, Tuple [str, str]], Dict[str, int]]:
    
    variables:Dict[str, Tuple [str, str]] = {}
    if env_config.get('ep_environment_variables', False):
        variables = {variable: (variable, 'Environment') for variable in env_config['ep_environment_variables']}
    
    var_handles: Dict[str, int] = {}
    
    return variables, var_handles

def thermal_zone_variables(env_config: Dict[str, Any], _thermal_zone_ids:Set) -> Tuple[Dict[str, Dict[str, Tuple[str,str]]],Dict[str,int]]:
    thermal_zone_variables: Dict[str, Tuple [str, str]] = {thermal_zone: {} for thermal_zone in _thermal_zone_ids}
    if env_config.get('ep_thermal_zones_variables', False):
        for thermal_zone in _thermal_zone_ids:
            thermal_zone_variables[thermal_zone].update({variable: (variable, thermal_zone) for variable in env_config['ep_thermal_zones_variables']})
    thermal_zone_var_handles: Dict[str, int] = {thermal_zone: {} for thermal_zone in _thermal_zone_ids}
    
    return thermal_zone_variables, thermal_zone_var_handles

def object_variables(env_config: Dict[str, Any], _thermal_zone_ids: Set) -> Tuple[Dict[str, Dict[str, Tuple[str,str]]],Dict[str,int]]:
    object_variables: Dict[str, Dict[str, Tuple [str, str]]] = {}
    if env_config.get('ep_object_variables', False):
        object_variables = env_config['ep_object_variables']
    object_var_handles = {thermal_zone: {} for thermal_zone in _thermal_zone_ids}
    
    return object_variables, object_var_handles

def meters(env_config: Dict[str, Any]) -> Tuple[Dict[str, str], Dict[str,int]]:
    meters: Dict[str,str] = {}
    if env_config.get('ep_meters', False):
        meters = {key: key for key in env_config['ep_meters']}
    meter_handles: Dict[str, int] = {}
    
    return meters, meter_handles

def actuators(env_config: Dict[str, Any], _agent_ids:Set) -> Tuple[Dict[str,Tuple[str,str,str]], Dict[str,int]]:
    
    actuators: Dict[str,Tuple[str,str,str]] = {agent: env_config['agents_config'][agent]['ep_actuator_config'] for agent in _agent_ids}
    actuator_handles: Dict[str, int] = {}
    
    return actuators, actuator_handles


"""This module is used to load an IDF file and recognize the content
of the file related with the agents that would be used in the eprllib
library.
"""

# Se debe cargar el path al archivo IDF y en el identificar los posibles
# actuadores, variables y metricas.

# Para esto se debe ejecutar un paso de tiempo la siulacion del IDF para
# generar los diferetnets tipos de archivos que EnergyPlus produce para
# verificar la configuracion de la simulacion.

# A partir de estas identificaciones, se podran establecer los espacios
# de observacion y accion, como asi tambien los agentes del entorno.

class EnergyPlusInputFile:

    def __initi__(self, idf_path:str, output_folder:str):
        # Se definen las variables de entrada
        self.path = idf_path
        self.output_folder = output_folder
        
        # Se establecen variables que almacenaran las propiedades del IDF
        self.variables = None
        self.meters = None
        self.actuators = None

        # Se inicia un paso de tiempo de simulacion en E+ para determinar las propiedades del IDF
        self.start()

    def start(self):
        # Se establece un estado para E+ API
        self.state_argument = api.state_managment.init()
        # Se establece un punto de llamado para detener la simulacion en el primer paso de tiempo

        # Se establece la salida de la consola como False

        # Se genera el string de comando para iniiar la simulacion
        # Aqui se establece una propiedad de la clase que almacena la ubicacion de los archivos generados por E+
        
        # se inicia la simulacion




class User:
    
    def __init__(
        self,
        config: dict,
    ):
        self.config = config
        
    def action(self):
        return
    
    def precense_schedule(self):
        """Perform and precense schedule for the user in the house.
        """
        return
    
    def temperature_schedule(self, hora:int):
        """Esta función establece las temperaturas del termostato de una vivienda según la hora del
        día.

        Args:
            obs (dict): El diccionario debe contener en su observación al menos los siguientes
            elementos:
                'hora' es la hora del día de 0 a 23

        Returns:
            int: Temperatura de calefacción
            int: Temperatura de refrigeración
        """
        if hora <= 7 or hora >= 23:
            Theat = 17
            Tchill = 28
        
        elif hora > 7 and hora < 23:
            Theat = 20
            Tchill = 25

        return Theat, Tchill


"""This file provide the entire parameters configuration of the 
multi-agent environment approach.
"""
from eprllib.tools.rewards import dalamagkidis_2007

env_config = {
    # == FILES DIRECTORUES ==
    # The path to the EnergyPlus model in the format of epJSON file.
    'epjson': 'path/to/epjson_file.json',
    # The path to the EnergyPlus weather file in the format of epw file.
    'epw': 'path/to/epw_file.epw',
    # The path to the output directory for the EnergyPlus simulation.
    'output': 'path/to/output_directory',
    
    # == ENVIRONMENT CONFIGURATION ==
    # Dictionary of tuples to define the Agents. The Dict has the following format: 
    #   Dict[str, Tuple[Tuple[str, str, str], str, int]]
    # In terms of meaning of each variable of the agent definition, we can say:
    #   Dict['agent_name', Tuple[Tuple['EP_type_of_actuator', 'EP_variable', 'EP_name or EP_zone'], 'EP_zone', 'eprllib_type']].
    # where:
    #   'agent_name'(str): The unique name assign to the agent by the user.
    #   'EP_type_of_actuator'(str): EnergyPlus provide with some actuators in her Python API. Here goes the name of the type 
    # of actuator to control.
    #   'EP_variable'(str): EnergyPlus provide with some actuators in her Python API. Here goes the name of the variable to modify.
    #   'EP_name or EP_zone'(str): EnergyPlus provide with some actuators in her Python API. Some of them use an unique name to 
    # work and others the thermal zone. Here goes this idetifying name.
    #   'EP_zone'(str): The thermal zone where the agent is acting. For now, each agent only is capable to act in one thermal zone.
    # Note that each actuator is an agent in the environment. You can check the documentation of EnergyPlus to get more 
    # information about the actuators.
    #   'eprllib_type'(int): The eprllib_type(s) of the actuators follow the next correspondency of names:
    #       1: Cooling set point
    #       2: Heating set point
    #       3: Acondicionated Air Flow Rate
    #       4: North Window Opening
    #       5: East Window Opening
    #       6: South Window Opening
    #       7: West Window Opening
    #       8: North Window Shading
    #       9: East Window Shading
    #       10: South Window Shading
    #       11: West Window Shading
    #       12: Fan Flow Rate
    'agents_config': {
        'AirMassFlowRate Living': {
            'ep_actuator_config': ("Ideal Loads Air System", "Air Mass Flow Rate", "Thermal Zone: Living Ideal Loads Air System"),
            'thermal_zone': 'Thermal Zone: Living',
            'actuator_type': 3,
            'agent_indicator': 1,
        },
    # The action space for all the agents.
    'action_space': 'gym.spaces.Discrete() type',
    # Definition of the observation space.
    # define if the actuator state will be used as an observation for the agent.
    'use_actuator_state': True,
    # define if agent indicator will be used as an observation for the agent. This is recommended 
    # True for muilti-agent usage and False for single agent case.
    'use_agent_indicator': True,
    # define if the agent/actuator type will be used. This is recommended for different types of 
    # agents actuating in the same environment.
    'use_agent_type': True,
    # define if the building properties will be used as an observation for the agent. This is recommended
    # if different buildings will be used with the same policy.
    'use_building_properties': True,
    # The episode config define important aspects about the building to be simulated
    # in the episode. All this parameters are mandatory.
    'building_properties': {
        'ThermalZone': {
            # Net Conditioned Building Area [m2]
            'building_area': 20.75,
            'aspect_ratio': 1.35,
            # Conditioned Window-Wall Ratio, Gross Window-Wall Ratio [%]
            'window_area_relation_north': 56.67,
            'window_area_relation_east': 18.19,
            'window_area_relation_south': 2.59,
            'window_area_relation_west': 0,
            'inercial_mass': "auto",
            'construction_u_factor': "auto",
            # User-Specified Maximum Total Cooling Capacity [W]
            'E_cool_ref': 2500,
            # User-Specified Maximum Sensible Heating Capacity [W]
            'E_heat_ref': 2500,
        },
    },
    # We use the internal variables of EnergyPlus to provide with a prediction of the weather
    # time ahead.
    # The variables to predict are:
    #   - Dry Bulb Temperature in °C with squer desviation of 2.05 °C, 
    #   - Relative Humidity in % with squer desviation of 20%, 
    #   - Wind Direction in degree with squer desviation of 40°, 
    #   - Wind Speed in m/s with squer desviation of 3.41 m/s, 
    #   - Barometric pressure in Pa with a standart deviation of 1000 Pa, 
    #   - Liquid Precipitation Depth in mm with desviation of 0.5 mm.
    # This are predicted from the next hour into the 24 hours ahead defined.
    'use_one_day_weather_prediction': True,
    # Dictionary of tuples with the EnergyPlus variables to observe and their corresponding names.
    # The tuples has the format of (name_of_the_variable, Thermal_Zone_name). You can check the
    # EnergyPlus output file to get the names of the variables.
    'ep_environment_variables': [
        'Site Outdoor Air Drybulb Temperature',
    ],
    'ep_thermal_zones_variables': [
        'Zone Mean Air Temperature',
        'Zone Air Relative Humidity',
    ],
    'ep_object_variables': [
        ('Zone Thermal Comfort Fanger Model PPD', 'Living Occupancy'),
    ],
    # Dictionary of names of meters from EnergyPlus to observe.
    'ep_meters': [
        'Electricity:Zone:THERMAL ZONE: LIVING',
        'NaturalGas:Zone:THERMAL ZONE: LIVING',
        'Heating:DistrictHeatingWater',
        'Cooling:DistrictCooling',
    ],
    # The time variables to observe in the EnergyPlus simulation. The format is a list of
    # the names described in the EnergyPlus epJSON format documentation 
    # (https://energyplus.readthedocs.io/en/latest/schema.html) related with
    # temporal variables. All the options are listed bellow.
    'time_variables': [
        # Gets a simple sum of the values of the date/time function. Could be used in 
        # random seeding.
        'actual_date_time',
        # Gets a simple sum of the values of the time part of the date/time function. 
        # Could be used in random seeding.
        'actual_time',
        # Get the current time of day in hours, where current time represents the end 
        # time of the current time step.
        'current_time',
        # Get the current day of month (1-31)
        'day_of_month',
        # Get the current day of the week (1-7) 
        'day_of_week',
        # Get the current day of the year (1-366)
        'day_of_year',
        # Gets a flag for the current day holiday type: 0 is no holiday, 1 is holiday 
        # type #1, etc.
        'holiday_index',
        # Get the current hour of the simulation (0-23)
        'hour',
        # Get the current minutes into the hour (1-60)
        'minutes',
        # Get the current month of the simulation (1-12)
        'month',
        # Returns the number of zone time steps in an hour, which is currently a 
        # constant value throughout a simulation.
        'num_time_steps_in_hour',
        # Gets the current system time step value in EnergyPlus. The system time 
        # step is variable and fluctuates during the simulation.
        'system_time_step',
        # Get the “current” year of the simulation, read from the EPW. All simulations 
        # operate at a real year, either user specified or automatically selected by 
        # EnergyPlus based on other data (start day of week + leap year option).
        'year',
        # Gets the current zone time step value in EnergyPlus. The zone time step is 
        # variable and fluctuates during the simulation.
        'zone_time_step',
        # The current zone time step index, from 1 to the number of zone time steps 
        # per hour.
        'zone_time_step_number',
    ],
    # The weather variables are related with weather values in the present timestep for the
    # agent. The following list provide all the options avialable.
    # To weather predictions see the 'weather_prob_days' config that is follow in this file.
    'weather_variables': [
        # Gets a flag for whether the it is currently raining. The C API returns an 
        # integer where 1 is yes and 0 is no, this simply wraps that with a bool 
        # conversion.
        'is_raining',
        # Gets a flag for whether the sun is currently up. The C API returns an integer 
        # where 1 is yes and 0 is no, this simply wraps that with a bool conversion.
        'sun_is_up',
        # Gets the specified weather data at the specified hour and time step index 
        # within that hour.
        'today_weather_albedo_at_time',
        'today_weather_beam_solar_at_time',
        'today_weather_diffuse_solar_at_time',
        'today_weather_horizontal_ir_at_time',
        'today_weather_is_raining_at_time',
        'today_weather_is_snowing_at_time',
        'today_weather_liquid_precipitation_at_time',
        'today_weather_outdoor_barometric_pressure_at_time',
        'today_weather_outdoor_dew_point_at_time',
        'today_weather_outdoor_dry_bulb_at_time',
        'today_weather_outdoor_relative_humidity_at_time',
        'today_weather_sky_temperature_at_time',
        'today_weather_wind_direction_at_time',
        'today_weather_wind_speed_at_time',
        'tomorrow_weather_albedo_at_time',
        'tomorrow_weather_beam_solar_at_time',
        'tomorrow_weather_diffuse_solar_at_time',
        'tomorrow_weather_horizontal_ir_at_time',
        'tomorrow_weather_is_raining_at_time',
        'tomorrow_weather_is_snowing_at_time',
        'tomorrow_weather_liquid_precipitation_at_time',
        'tomorrow_weather_outdoor_barometric_pressure_at_time',
        'tomorrow_weather_outdoor_dew_point_at_time',
        'tomorrow_weather_outdoor_dry_bulb_at_time',
        'tomorrow_weather_outdoor_relative_humidity_at_time',
        'tomorrow_weather_sky_temperature_at_time',
        'tomorrow_weather_wind_direction_at_time',
        'tomorrow_weather_wind_speed_at_time',
    ],
    # The information variables are important to provide information for the reward
    # function. The observation is pass trough the agent as a NDArray but the info
    # is a dictionary. In this way, we can identify clearly the value of a variable
    # with the key name. All the variables used in the reward function must to be in
    # the infos_variables list. The name of the variables must to corresponde with the
    # names defined in the earlier lists. 
    'infos_variables': {
        'Thermal Zone': [
            'variable name',
        ],
    },
    # There are occasions where some variables are consulted to use in training but are
    # not part of the observation space. For that variables, you can use the following 
    # list. An strategy, for example, to use the Fanger PPD value in the reward function
    # but not in the observation space is to aggregate the PPD into the 'infos_variables' and
    # in the 'no_observable_variables' list.
    "no_observable_variables": [],
    # This method define the properties of the episode, taking the env_config dict and returning it
    # with modifications.
    'episode_config_fn': None,
    'episode_config': None,
    # Sometimes is useful to cut the simulation RunPeriod into diferent episodes. By default, 
    # an episode is a entire RunPeriod EnergyPlus simulation. If you set the 'cut_episode_len'
    # in 1 (day) you will truncate the, for example, annual simulation into 365 episodes.
    'cut_episode_len': None,
    # timeout define the time that the environment wait for an observation and the time that
    # the environment wait to apply an action in the EnergyPlus simulation. After that time,
    # the episode is finished. If your environment is time consuming, you can increase this
    # limit. By default the value is 10 seconds.
    "timeout": 10,
    # The reward funtion take the arguments EnvObject (the GymEnv class) and the infos dictionary.
    # As a return, gives a float number as reward. See eprllib.tools.rewards
    'reward_function': dalamagkidis_2007,
    # Reward function config dictionary
    'reward_function_config': {
        'Thermal zone': {
            'cut_reward_len_timesteps': 1,
            'comfort_reward': True,
            'energy_reward': True,
            'co2_reward': True,
            'w1': 0.80,
            'w2': 0.01,
            'w3': 0.20,
            'energy_ref': 6805274,
            'co2_ref': 870,
            'occupancy_name': 'occupancy',
            'ppd_name': 'ppd',
            'T_interior_name': 'Ti',
            'cooling_name': 'cooling',
            'heating_name': 'heating',
            'co2_name': 'co2',
        },
    },
    # For dubugging is better to print in the terminal the outputs of the EnergyPlus simulation process.
    'ep_terminal_output': False,
    # The action space is defined as a continuous space between 0 and 1.
    # In general, EnergyPlus actuator need to be feed with a diferent scale of numbers, e.g. temperature between
    # 16 and 27 °C for the thermostat. In this way, we don't use actions from 0 to n directly in the EnergyPlus simulation. With
    # the objective to transform appropiately the continuous action into a value for the actuator, we
    # define the eprllib.Tools.ActionTransformers.ActionTransformer class.
    'action_transformer': None,
}


"""This script will be contain some action transformer methods to implement in
eprllib. Most of them are applied in the test section where examples to test the 
library are developed.
"""
from typing import Dict, Set, Any

class ActionTransformer:
    def __init__(
        self,
        agents_config: Dict,
        _agent_ids: Set,
    ):
        self.agents_config = agents_config
        self._agent_ids = _agent_ids
    
    def transform_action(self, action:Dict[str,float]) -> Dict[str, Any]:
        raise NotImplementedError

"""Example"""

class DualSetPointThermostat(ActionTransformer):
    def __init__(self, agents_config:Dict, _agent_ids:Set):
        super().__init__(agents_config, _agent_ids)
    
    def transform_action(self, action:Dict[str,int]) -> Dict[str, float|int]:
        action_transformed = {agent: 0. for agent in self._agent_ids}
        for agent in self._agent_ids:
            actuator_type = self.agents_config[agent]['actuator_type']
            if actuator_type == 1: # Cooling
                # transfort value between [0,1] to [23,27]
                action_transformed[agent] = 23 + (action[agent]/10)*(27-23)
            elif actuator_type == 2: # Heating
                # transfort value between [0,1] to [18,22]
                action_transformed[agent] = 18 + (action[agent]/10)*(22-18)
            elif actuator_type == 3: # Flow Rate
                action_transformed[agent] = action[agent]/20
            else:
                raise ValueError('Actuator type not valid.')
        
        return action_transformed


"""Utilities and methods to configurate the execution of the episode in EnergyPlus with RLlib.
"""
import numpy as np
import os
import json
from typing import List, Dict

def random_building_config(env_config:dict):
    """This method define the path to the epJSON file.

    Args:
        env_config (dict): Environment configuration.

    Return:
        dict: The method returns the env_config with modifications.
        
    Example:
    
        env_config = {
            ...
            'episode_config': {
                'epjson_files_folder_path': 'path/to/epjson/files',
                'id_epjson_file': None
            }
        }
    """
    if env_config.get('epjson', False) == False:
        env_config['epjson'] = ''
    
    # Define the properties of the method
    epjson_files_folder_path: str = None # This is mandatory
    id_epjson_file: int = None # The default is a random int, but the user can indicated a diferent number
    
    # For each property, assign the default or the user input
    epjson_files_folder_path = env_config['episode_config'].get('epjson_files_folder_path', None)
    if epjson_files_folder_path is None:
        ValueError('epjson_files_folder_path is not defined')
    
    id_epjson_file = env_config['episode_config'].get('id_epjson_file', None)
    if id_epjson_file is None:
        id_epjson_file = np.random.randint(0, len(os.listdir(epjson_files_folder_path)))
    
    # The path to the epjson file is defined
    env_config['epjson'] = os.path.join(epjson_files_folder_path, os.listdir(epjson_files_folder_path)[id_epjson_file])
    
    return env_config

def random_weather_config(env_config:Dict) -> Dict:
    """This method define the path to the epJSON file.

    Args:
        env_config (dict): Environment configuration.

    Return:
        Dict: The method returns the epw path.
        
    Example:
    
        env_config = {
            ...
            'episode_config': {
                'epw_files_folder_path': 'path/to/epw/files',
                'id_epw_file': None
            }
        }
    """
    if env_config.get('epw', False) == False:
        env_config['epw'] = ''
    
    # Define the properties of the method
    epw_files_folder_path: str = None # This is mandatory
    id_epw_file: int = None # The default is a random int, but the user can indicated a diferent number
    
    # For each property, assign the default or the user input
    epw_files_folder_path = env_config['episode_config'].get('epw_files_folder_path', None)
    if epw_files_folder_path is None:
        ValueError('epw_files_folder_path is not defined')
    
    id_epw_file = env_config['episode_config'].get('id_epw_file', None)
    if id_epw_file is None:
        id_epw_file = np.random.randint(0, len(os.listdir(epw_files_folder_path)))
    
    # The path to the epjson file is defined
    env_config['epw'] = os.path.join(epw_files_folder_path, os.listdir(epw_files_folder_path)[id_epw_file])
    
    return env_config


def episode_epJSON(env_config:Dict) -> Dict:
    """This method define the properties of the episode. Changing some properties as weather or 
    Run Time Period, and defining others fix properties as volumen or window area relation.
    
    Args:
        env_config (dict): Environment configuration.
        
    Return:
        dict: The method returns the env_config with modifications.
    """
    # If the path to epjson is not set, arraise a error.
    if not env_config['episode_config'].get('epjson', False):
        raise ValueError('epjson is not defined')
    
    # Establish the epJSON Object, it will be manipulated to modify the building model.
    with open(env_config['episode_config']['epjson']) as file:
        epJSON_object: dict = json.load(file)
    
    # == BUILDING ==
    # The building volume is V=h(high)*w(weiht)*l(large) m3
    h = 3
    w = (10-3) * np.random.random_sample() + 3
    l = (10-3) * np.random.random_sample() + 3
    
    # Calculate the aspect relation
    env_config['building_properties']['building_area'] = w*l
    env_config['building_properties']['aspect_ratio'] = w/l
    
    # Change the dimension of the building and windows
    window_area_relation = []
    env_config['building_properties']['window_area_relation_north'] = (0.9-0.05) * np.random.random_sample() + 0.05
    env_config['building_properties']['window_area_relation_east'] = (0.9-0.05) * np.random.random_sample() + 0.05
    env_config['building_properties']['window_area_relation_south'] = (0.9-0.05) * np.random.random_sample() + 0.05
    env_config['building_properties']['window_area_relation_west'] = (0.9-0.05) * np.random.random_sample() + 0.05
    window_area_relation.append(env_config['building_properties']['window_area_relation_north'])
    window_area_relation.append(env_config['building_properties']['window_area_relation_east'])
    window_area_relation.append(env_config['building_properties']['window_area_relation_south'])
    window_area_relation.append(env_config['building_properties']['window_area_relation_west'])
    window_area_relation = np.array(window_area_relation)
    building_dimension(epJSON_object, h, w, l,window_area_relation)
    
    # Define the type of construction (construction properties for each three layers)
    # Walls
    # Exterior Finish
    # Se toman las propiedades en el rango de valores utilizados en BOpt
    epJSON_object["Material"]["wall_exterior"]["thickness"] = (0.1017-0.0095) * np.random.random_sample() + 0.0095
    epJSON_object["Material"]["wall_exterior"]["conductivity"] = (0.7934-0.0894) * np.random.random_sample() + 0.0894
    epJSON_object["Material"]["wall_exterior"]["density"] = (1762.3-177.822) * np.random.random_sample() + 177.822
    epJSON_object["Material"]["wall_exterior"]["specific_heat"] = (1172.37-795.53) * np.random.random_sample() + 795.53
    epJSON_object["Material"]["wall_exterior"]["thermal_absorptance"] = (0.97-0.82) * np.random.random_sample() + 0.82
    epJSON_object["Material"]["wall_exterior"]["solar_absorptance"] = epJSON_object["Material"]["wall_exterior"]["visible_absorptance"] = (0.89-0.3) * np.random.random_sample() + 0.3
    # Inter layers
    # Se realizó un equivalente de los diferentes materiales utilizados en BOpt para resumirlos en una sola capa equivalente.
    epJSON_object["Material"]["wall_inter"]["thickness"] = (0.29-0.1524) * np.random.random_sample() + 0.1524
    epJSON_object["Material"]["wall_inter"]["conductivity"] = (0.8656-0.0474) * np.random.random_sample() + 0.0474
    epJSON_object["Material"]["wall_inter"]["density"] = (1822.35-84.57) * np.random.random_sample() + 84.57
    epJSON_object["Material"]["wall_inter"]["specific_heat"] = (1214.24-852.57) * np.random.random_sample() + 852.57
    epJSON_object["Material"]["wall_inter"]["thermal_absorptance"] = 0.9
    epJSON_object["Material"]["wall_inter"]["solar_absorptance"] = epJSON_object["Material"]["wall_inter"]["visible_absorptance"] = (0.71-0.65) * np.random.random_sample() + 0.65
    
    # Ceiling/Roof (exterior, inter layers)
    # Exterior Finish
    epJSON_object["Material"]["roof_exterior"]["thickness"] = (0.01906-0.0005) * np.random.random_sample() + 0.0005
    epJSON_object["Material"]["roof_exterior"]["conductivity"] = (50.04-0.1627) * np.random.random_sample() + 0.1627
    epJSON_object["Material"]["roof_exterior"]["density"] = (7801.75-1121.4) * np.random.random_sample() + 1121.4
    epJSON_object["Material"]["roof_exterior"]["specific_heat"] = (1465.46-460.57) * np.random.random_sample() + 460.57
    epJSON_object["Material"]["roof_exterior"]["thermal_absorptance"] = (0.95-0.88) * np.random.random_sample() + 0.88
    epJSON_object["Material"]["roof_exterior"]["solar_absorptance"] = epJSON_object["Material"]["roof_exterior"]["visible_absorptance"] = (0.93-0.6) * np.random.random_sample() + 0.6
    # Inter layers
    epJSON_object["Material"]["roof_inner"]["thickness"] = (0.1666-0.0936) * np.random.random_sample() + 0.0936
    epJSON_object["Material"]["roof_inner"]["conductivity"] = 0.029427
    epJSON_object["Material"]["roof_inner"]["density"] = 32.04
    epJSON_object["Material"]["roof_inner"]["specific_heat"] = 1214.23
    epJSON_object["Material"]["roof_inner"]["thermal_absorptance"] = 0.9
    epJSON_object["Material"]["roof_inner"]["solar_absorptance"] = epJSON_object["Material"]["roof_inner"]["visible_absorptance"] = 0.7

    # Internal Mass layer (interior layer)
    epJSON_object["Material"]["wall_inner"]["thickness"] = epJSON_object["Material"]["roof_inner"]["thickness"] = (0.0318-0.0127) * np.random.random_sample() + 0.0127
    epJSON_object["Material"]["wall_inner"]["conductivity"] = epJSON_object["Material"]["roof_inner"]["conductivity"] = 0.1602906
    epJSON_object["Material"]["wall_inner"]["density"] = epJSON_object["Material"]["roof_inner"]["density"] = 801
    epJSON_object["Material"]["wall_inner"]["specific_heat"] = epJSON_object["Material"]["roof_inner"]["specific_heat"] = 837.4
    epJSON_object["Material"]["wall_inner"]["thermal_absorptance"] = epJSON_object["Material"]["roof_inner"]["thermal_absorptance"] = 0.9
    epJSON_object["Material"]["wall_inner"]["solar_absorptance"] = epJSON_object["Material"]["wall_inner"]["visible_absorptance"] = epJSON_object["Material"]["roof_inner"]["solar_absorptance"] = epJSON_object["Material"]["roof_inner"]["visible_absorptance"] = 0.6

    # Windows
    # Change the window thermal properties
    epJSON_object['WindowMaterial:SimpleGlazingSystem']['WindowMaterial']['u_factor'] = (4.32-0.9652) * np.random.random_sample() + 0.9652
    epJSON_object['WindowMaterial:SimpleGlazingSystem']['WindowMaterial']['solar_heat_gain_coefficient'] = (0.67-0.26) * np.random.random_sample() + 0.26
    
    # The internal thermal mass is modified.
    for key in [key for key in epJSON_object["InternalMass"].keys()]:
        epJSON_object["InternalMass"][key]["surface_area"] = np.random.randint(10,40)
    
    # The total inertial thermal mass is calculated.
    env_config['building_properties']['inercial_mass'] = inertial_mass(epJSON_object)
    
    # The global U factor is calculated.
    env_config['building_properties']['construction_u_factor'] = u_factor(epJSON_object)
    
    # The limit capacity of bouth cooling and heating are changed.
    E_cool_ref = env_config['building_properties']['E_cool_ref'] = (3000 - 100)*np.random.random_sample() + 100
    E_heat_ref = env_config['building_properties']['E_heat_ref'] = (3000 - 100)*np.random.random_sample() + 100
    HVAC_names = [key for key in epJSON_object["ZoneHVAC:IdealLoadsAirSystem"].keys()]
    number_of_timesteps_per_hour = epJSON_object['Timestep']['Timestep 1']['number_of_timesteps_per_hour']
    for hvac in range(len(HVAC_names)):
        epJSON_object["ZoneHVAC:IdealLoadsAirSystem"][HVAC_names[hvac]]["maximum_sensible_heating_capacity"] = E_heat_ref
        epJSON_object["ZoneHVAC:IdealLoadsAirSystem"][HVAC_names[hvac]]["maximum_total_cooling_capacity"] = E_cool_ref
    if env_config.get('reward_function_config', False):
        if env_config['reward_function_config'].get('cooling_energy_ref', False):
            env_config['reward_function_config']['cooling_energy_ref'] = E_cool_ref*number_of_timesteps_per_hour*3600
        if env_config['reward_function_config'].get('heating_energy_ref', False):
            env_config['reward_function_config']['heating_energy_ref'] = E_heat_ref*number_of_timesteps_per_hour*3600
      
            
    # Select the schedule file for loads
    env_config = random_weather_config(env_config)
    
    # The new modify epjson file is writed in the results folder created by RLlib
    # If don't exist, reate a folder call 'models' into env_config['episode_config']['epjson_files_folder_path']
    if not os.path.exists(f"{env_config['episode_config']['epjson_files_folder_path']}/models"):
        os.makedirs(f"{env_config['episode_config']['epjson_files_folder_path']}/models")
    
    env_config["epjson"] = f"{env_config['episode_config']['epjson_files_folder_path']}/models/model-{env_config['episode']:08}-{os.getpid():05}.epJSON"
    with open(env_config["epjson"], 'w') as fp:
        json.dump(epJSON_object, fp, sort_keys=False, indent=4)
        # The new modify epjson file is writed.
    
    return env_config

def inertial_mass_calculation(env_config:Dict) -> float:
    """The function reads the epjson file path from the env_config dictionary. If the epjson
    key is not present, it raises a ValueError. Otherwise, it opens the specified epjson
    file, loads its contents as a Python dictionary (epJSON_object), and then calls the 
    inertial_mass function (not shown in the provided code) with epJSON_object
    as an argument. The result of inertial_mass is returned as a float value.

    Args:
        env_config (dict): A dictionary containing the environment configuration, including the 
            episode_config key, which should have an epjson key with the path to the 
            EnergyPlus JSON file.

    Returns:
        float: The inertial mass calculated based on the contents of the epjson file.
    
    Example:
        env_config = {
            'episode_config': {
                'epjson': 'path/to/epjson_file.json'
            }
        }

        inertial_mass_value = inertial_mass_calculation(env_config)
        print(f"Inertial mass: {inertial_mass_value}")

    """
    # If the path to epjson is not set, arraise a error.
    if not env_config.get('epjson', False):
        raise ValueError('epjson is not defined')
    
    with open(env_config['epjson']) as file:
        epJSON_object: dict = json.load(file)

    return inertial_mass(epJSON_object)

def inertial_mass(epJSON_object:Dict[str,Dict]) -> float:
    """The inertial_mass function calculates the total thermal mass of a building based 
    on the construction materials and surface areas specified in an EnergyPlus JSON 
    object.

    Args:
        epJSON_object (dict[str,dict]): A dictionary containing information about the 
        building surfaces, materials, and constructions in the EnergyPlus JSON format. 
        The keys of the outer dictionary represent different object types (e.g., 
        "BuildingSurface:Detailed", "Material", "Construction"), and the values are 
        dictionaries containing the properties of each object instance.

    Returns:
        float: The total thermal mass of the building in J/°C (Joules per degree Celsius). 
        The thermal mass represents the amount of heat energy required to raise the 
        temperature of the building's construction materials by one degree Celsius.
    """
    # se define una lista para almacenar
    masas_termicas = []
    
    building_surfaces = []
    if epJSON_object.get("BuildingSurface:Detailed", False):
        building_surfaces = [key for key in epJSON_object["BuildingSurface:Detailed"].keys()]
    # se obtienen los nombres de las superficies de la envolvente
    internal_mass_surfaces = []
    if epJSON_object.get("InternalMass", False):
        internal_mass_surfaces = [key for key in epJSON_object["InternalMass"].keys()]
    
    all_building_keys = [key for key in epJSON_object.keys()]
    all_material_list = ['Material','Material:NoMass','Material:InfraredTransparent','Material:AirGap',
        'Material:RoofVegetation','WindowMaterial:SimpleGlazingSystem','WindowMaterial:Glazing',
        'WindowMaterial:GlazingGroup:Thermochromic','WindowMaterial:Glazing:RefractionExtinctionMethod',
        'WindowMaterial:Gas','WindowGap:SupportPillar','WindowGap:DeflectionState',
        'WindowMaterial:GasMixture','WindowMaterial:Gap'
    ]
    materials_dict = {}
    for material in all_material_list:
        if material in all_building_keys:
            materials_dict[material] = epJSON_object[material].keys()
    # se obtienen los nombres de los diferentes tipos de materiales
    
    # lazo para consultar cada superficie de la envolvente
    for surface in building_surfaces:
        # se calcula el área de la superficie
        area = material_area(epJSON_object,surface)
        # se identifica la consutrucción
        s_construction = epJSON_object['BuildingSurface:Detailed'][surface]['construction_name']
        
        # se obtiene la densidad del materia: \rho[kg/m3]
        # se obtiene el calor específico del material: C[J/kg°C]
        # se calcula el volumen que ocupa el material: V[m3]=area*thickness
        # se calcula la masa térmica: M[J/°C] = \rho[kg/m3] * C[J/kg°C] * V[m3]
        
        # se establece un lazo para calcular la masa térmica de cada capa
        m_surface = 0
        layers = [key for key in epJSON_object['Construction'][s_construction].keys()]
        for layer in layers:
            material = epJSON_object['Construction'][s_construction][layer]
            material_list = find_dict_key_by_nested_key(
                material,
                materials_dict
            )
            # se obtiene el espesor y la conductividad térmica del material de la capa
            if material_list == 'Material:NoMass' or material_list == 'Material:AirGap' or material_list == 'Material:InfraredTransparent' or material_list == 'WindowMaterial:Gas':
                m_capa = 0
            else:
                espesor_capa = epJSON_object[material_list][material]['thickness']
                calor_especifico_capa = epJSON_object[material_list][material]['specific_heat']
                densidad_capa = epJSON_object[material_list][material]['density']
                m_capa = area * espesor_capa * calor_especifico_capa * densidad_capa

            # se suma la resistencia de la superficie
            m_surface += m_capa
        # se guarda la resistencia de la superficie
        masas_termicas.append(m_surface)
    
    # se suma la masa interna asignada
    for surface in internal_mass_surfaces:
        # se calcula el área de la superficie
        area = epJSON_object['InternalMass'][surface]['surface_area']
        # se identifica la consutrucción
        s_construction = epJSON_object['InternalMass'][surface]['construction_name']
        
        # se obtiene la densidad del materia: \rho[kg/m3]
        # se obtiene el calor específico del material: C[J/kg°C]
        # se calcula el volumen que ocupa el material: V[m3]=area*thickness
        # se calcula la masa térmica: M[J/°C] = \rho[kg/m3] * C[J/kg°C] * V[m3]
        
        # se establece un lazo para calcular la masa térmica de cada capa
        m_surface = 0
        layers = [key for key in epJSON_object['Construction'][s_construction].keys()]
        for layer in layers:
            material = epJSON_object['Construction'][s_construction][layer]
            material_list = find_dict_key_by_nested_key(
                material,
                materials_dict
            )
            # se obtiene el espesor y la conductividad térmica del material de la capa
            if material_list == 'Material:NoMass' or material_list == 'Material:AirGap' or material_list == 'Material:InfraredTransparent' or material_list == 'WindowMaterial:Gas':
                m_capa = 0
            else:
                espesor_capa = epJSON_object[material_list][material]['thickness']
                calor_especifico_capa = epJSON_object[material_list][material]['specific_heat']
                densidad_capa = epJSON_object[material_list][material]['density']
                m_capa = area * espesor_capa * calor_especifico_capa * densidad_capa

            # se suma la resistencia de la superficie
            m_surface += m_capa
        # se guarda la resistencia de la superficie
        masas_termicas.append(m_surface)
    
    # Cálculo de la masa termica total
    M_total = 0.
    for m in range(0,len(masas_termicas)-1,1):
        M_total += masas_termicas[m]
    
    return M_total

def u_factor_calculation(env_config:Dict) -> float:
    """The u_factor_calculation function calculates the U-factor (a measure of heat 
    transfer through a building element) based on the provided environment configuration 
    (env_config).

    Args:
        env_config (dict): A dictionary containing the environment configuration, including 
        the episode_config key, which should have an epjson key with the path to the 
        EnergyPlus JSON file.

    Returns:
        float: The calculated U-factor value.
        
    Example:
        env_config = {
            'episode_config': {
                'epjson': 'path/to/epjson_file.json'
            }
        }

        u_factor_value = u_factor_calculation(env_config)
        print(f"U-factor: {u_factor_value}")

    """
    # If the path to epjson is not set, arraise a error.
    if not env_config.get('epjson', False):
        raise ValueError('epjson is not defined')
    with open(env_config['epjson']) as file:
        epJSON_object: dict = json.load(file)
    #  Calculate the u_factor
    return u_factor(epJSON_object)

def u_factor(epJSON_object:Dict[str,Dict]) -> float:
    """This function select all the building surfaces and fenestration surfaces and calculate the
    global U-factor of the building, like EnergyPlus does.
    """
    # se define una lista para almacenar las resistencias de cada supercie
    resistences = []
    areas = []
    # se obtienen los nombres de las superficies de la envolvente
    building_surfaces = [key for key in epJSON_object['BuildingSurface:Detailed'].keys()]
    fenestration_surfaces = [key for key in epJSON_object['FenestrationSurface:Detailed'].keys()]
    # se obtienen los nombres de los diferentes tipos de materiales
    
    all_building_keys = [key for key in epJSON_object.keys()]
    all_material_list = ['Material','Material:NoMass','Material:InfraredTransparent','Material:AirGap',
        'Material:RoofVegetation','WindowMaterial:SimpleGlazingSystem','WindowMaterial:Glazing',
        'WindowMaterial:GlazingGroup:Thermochromic','WindowMaterial:Glazing:RefractionExtinctionMethod',
        'WindowMaterial:Gas','WindowGap:SupportPillar','WindowGap:DeflectionState',
        'WindowMaterial:GasMixture','WindowMaterial:Gap'
    ]
    materials_dict = {}
    for material in all_material_list:
        if material in all_building_keys:
            materials_dict[material] = epJSON_object[material].keys()
    # se obtienen los nombres de los diferentes tipos de materiales
    
    # lazo para consultar cada superficie de la envolvente
    for surface in building_surfaces:
        # se calcula el área de la superficie
        areas.append(material_area(epJSON_object,surface))
        # se identifica la consutrucción
        s_construction = epJSON_object['BuildingSurface:Detailed'][surface]['construction_name']
        # se establece un lazo para calcular la resistencia de cada capa
        r_surface = 0
        layers = [key for key in epJSON_object['Construction'][s_construction].keys()]
        for layer in layers:
            material = epJSON_object['Construction'][s_construction][layer]
            material_list = find_dict_key_by_nested_key(
                material,
                materials_dict
            )
            # se obtiene el espesor y la conductividad térmica del material de la capa
            if material_list == 'Material:NoMass' or material_list == 'Material:AirGap':
                r_capa = epJSON_object[material_list][material]['thermal_resistance']
            elif material_list == 'Material:InfraredTransparent':
                r_capa = 0
            elif material_list == 'WindowMaterial:Gas':
                espesor_capa = epJSON_object[material_list][material]['thickness']
                if epJSON_object[material_list][material]['gas_type'] == 'Air':
                    conductividad_capa = 0.0257
                elif epJSON_object[material_list][material]['gas_type'] == 'Argon':
                    conductividad_capa = 0.0162
                elif epJSON_object[material_list][material]['gas_type'] == 'Xenon':
                    conductividad_capa = 0.00576
                elif epJSON_object[material_list][material]['gas_type'] == 'Krypton':
                    conductividad_capa = 0.00943
                else:
                    print('El nombre del gas no corresponde con los que pueden utilizarse: Air, Argon, Xenon, Krypton.')
                    NameError
                r_capa = espesor_capa/conductividad_capa
            else:
                espesor_capa = epJSON_object[material_list][material]['thickness']
                conductividad_capa = epJSON_object[material_list][material]['conductivity']
                r_capa = espesor_capa/conductividad_capa

            # se suma la resistencia de la superficie
            r_surface += r_capa
        # se guarda la resistencia de la superficie
        resistences.append(r_surface)
        
    # lazo para consultar cada superfice de fenestración
    for fenestration in fenestration_surfaces:
        # se calcula el área de la superficie
        areas.append(fenestration_area(epJSON_object, fenestration))
        # se identifica la consutrucción
        s_construction = epJSON_object['FenestrationSurface:Detailed'][fenestration]['construction_name']
        # se establece un lazo para calcular la resistencia de cada capa
        r_surface = 0
        u_factor_windows = 0
        u_factor_window = 0
        layers = [key for key in epJSON_object['Construction'][s_construction].keys()]
        for layer in layers:
            material = epJSON_object['Construction'][s_construction][layer]
            material_list = find_dict_key_by_nested_key(
                material,
                materials_dict
            )
            # se obtiene el espesor y la conductividad térmica del material de la capa
            if material_list == 'Material:NoMass' or material_list == 'Material:AirGap':
                r_capa = epJSON_object[material_list][material]['thermal_resistance']
            elif material_list == 'Material:InfraredTransparent':
                r_capa = 0
            elif material_list == 'WindowMaterial:Gas':
                espesor_capa = epJSON_object[material_list][material]['thickness']
                if epJSON_object[material_list][material]['gas_type'] == 'Air':
                    conductividad_capa = 0.0257
                elif epJSON_object[material_list][material]['gas_type'] == 'Argon':
                    conductividad_capa = 0.0162
                elif epJSON_object[material_list][material]['gas_type'] == 'Xenon':
                    conductividad_capa = 0.00576
                elif epJSON_object[material_list][material]['gas_type'] == 'Krypton':
                    conductividad_capa = 0.00943
                else:
                    print('El nombre del gas no corresponde con los que pueden utilizarse: Air, Argon, Xenon, Krypton.')
                    NameError
                r_capa = espesor_capa/conductividad_capa
            elif material_list == 'WindowMaterial:SimpleGlazingSystem':
                u_factor_window = epJSON_object[material_list][material]['u_factor']
            else:
                espesor_capa = epJSON_object[material_list][material]['thickness']
                conductividad_capa = epJSON_object[material_list][material]['conductivity']
                r_capa = espesor_capa/conductividad_capa

            # se suma la resistencia de la superficie
            r_surface += r_capa
            u_factor_windows += u_factor_window
        # se guarda la resistencia de la superficie
        resistences.append(r_surface)

    # Cálculo de U-Factor en W/°C
    u_factor = 0.
    for n in range(0, len(areas)-1,1):
        u_factor =+ areas[n]/resistences[n]
    
    return u_factor + u_factor_windows

def material_area(epJSON_object, nombre_superficie):
    """_summary_

    Args:
        epJSON_object (_type_): _description_
        nombre_superficie (_type_): _description_

    Returns:
        _type_: _description_
    """
    # Calcula dos vectores que forman dos lados del cuadrilátero
    vector1 = [
        epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][1]['vertex_x_coordinate'] - epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][0]['vertex_x_coordinate'],
        epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][1]['vertex_y_coordinate'] - epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][0]['vertex_y_coordinate'],
        epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][1]['vertex_z_coordinate'] - epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][0]['vertex_z_coordinate']
    ]
    vector2 = [
        epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][2]['vertex_x_coordinate'] - epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][0]['vertex_x_coordinate'],
        epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][2]['vertex_y_coordinate'] - epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][0]['vertex_y_coordinate'],
        epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][2]['vertex_z_coordinate'] - epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][0]['vertex_z_coordinate']
    ]

    # Calcula el producto vectorial entre los dos vectores
    producto_vectorial = [
        vector1[1] * vector2[2] - vector1[2] * vector2[1],
        vector1[2] * vector2[0] - vector1[0] * vector2[2],
        vector1[0] * vector2[1] - vector1[1] * vector2[0]
    ]

    # Calcula el módulo del producto vectorial como el área del cuadrilátero
    area = 0.5 * (abs(producto_vectorial[0]) + abs(producto_vectorial[1]) + abs(producto_vectorial[2]))
    return area

def fenestration_area(epJSON_object, fenestration):
    """_summary_

    Args:
        epJSON_object (_type_): _description_
        fenestration (_type_): _description_

    Returns:
        _type_: _description_
    """
    
    # Calcula dos vectores que forman dos lados del cuadrilátero
    vector1 = [
        epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_2_x_coordinate'] - epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_1_x_coordinate'],
        epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_2_y_coordinate'] - epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_1_y_coordinate'],
        epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_2_z_coordinate'] - epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_1_z_coordinate']
    ]
    vector2 = [
        epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_3_x_coordinate'] - epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_1_x_coordinate'],
        epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_3_y_coordinate'] - epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_1_y_coordinate'],
        epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_3_z_coordinate'] - epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_1_z_coordinate']
    ]

    # Calcula el producto vectorial entre los dos vectores
    producto_vectorial = [
        vector1[1] * vector2[2] - vector1[2] * vector2[1],
        vector1[2] * vector2[0] - vector1[0] * vector2[2],
        vector1[0] * vector2[1] - vector1[1] * vector2[0]
    ]

    # Calcula el módulo del producto vectorial como el área del cuadrilátero
    area = 0.5 * (abs(producto_vectorial[0]) + abs(producto_vectorial[1]) + abs(producto_vectorial[2]))
    return area

def find_dict_key_by_nested_key(key, lists_dict):
    """_summary_

    Args:
        key (_type_): _description_
        lists_dict (_type_): _description_

    Returns:
        _type_: _description_
    """
    for dict_key, lst in lists_dict.items():
        if key in lst:
            return dict_key
    return None

def window_size_epJSON(epJSON_object, window:str, area_ventana:float):
    """Given a epJSON_object, return another epJSON_object with diferent size of windows.

    Args:
        epJSON_object (json): _description_
        window_name (str): _description_
        factor (float): _description_

    Returns:
        json: Devuelve el objeto epJSON modificado.
    """
    window_vertexs = {}
    # se extraen los valores de los vértices de cada ventana según el epJSON
    window_vertexs['v1x'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_1_x_coordinate']
    window_vertexs['v1y'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_1_y_coordinate']
    window_vertexs['v1z'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_1_z_coordinate']
    window_vertexs['v2x'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_2_x_coordinate']
    window_vertexs['v2y'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_2_y_coordinate']
    window_vertexs['v2z'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_2_z_coordinate']
    window_vertexs['v3x'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_3_x_coordinate']
    window_vertexs['v3y'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_3_y_coordinate']
    window_vertexs['v3z'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_3_z_coordinate']
    window_vertexs['v4x'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_4_x_coordinate']
    window_vertexs['v4y'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_4_y_coordinate']
    window_vertexs['v4z'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_4_z_coordinate']
    
    L = [] # agrupa los vertices en forma de lista: [[x1,y1,z1],[x2,y2,z2],[x3,y3,z3],[x4,y4,z4]]
    for l in range(1,5,1):
        vertex_x = 'v'+str(l)+'x'
        vertex_y = 'v'+str(l)+'y'
        vertex_z = 'v'+str(l)+'z'
        L.append([window_vertexs[vertex_x], window_vertexs[vertex_y], window_vertexs[vertex_z]])

    # Se calcula el factor de escala de la ventana
    area_ventana_old = fenestration_area(epJSON_object, window)
    factor_escala = area_ventana/area_ventana_old
    centro = calcular_centro(L)
    ventana_escalada = []
    for punto in L:
        nuevo_punto = [centro[0] + (punto[0] - centro[0]) * factor_escala**(1/2),
                    centro[1] + (punto[1] - centro[1]) * factor_escala**(1/2),
                    centro[2] + (punto[2] - centro[2]) * factor_escala**(1/2)]
        ventana_escalada.append(nuevo_punto)
    
    for l in range(1,5,1):
        epJSON_object['FenestrationSurface:Detailed'][window]['vertex_'+str(l)+'_x_coordinate'] = ventana_escalada[l-1][0]
        epJSON_object['FenestrationSurface:Detailed'][window]['vertex_'+str(l)+'_y_coordinate'] = ventana_escalada[l-1][1]
        epJSON_object['FenestrationSurface:Detailed'][window]['vertex_'+str(l)+'_z_coordinate'] = ventana_escalada[l-1][2]

def calcular_centro(ventana):
    # Calcula el centro de la ventana
    centro = [(ventana[0][0] + ventana[1][0] + ventana[2][0] + ventana[3][0]) / 4,
            (ventana[0][1] + ventana[1][1] + ventana[2][1] + ventana[3][1]) / 4,
            (ventana[0][2] + ventana[1][2] + ventana[2][2] + ventana[3][2]) / 4]
    return centro


def building_dimension(epJSON_object, h:float, w:float, l:float, window_area_relation:List):
    """_summary_

    Args:
        epJSON_object (_type_): _description_
        nombre_superficie (_type_): _description_

    Returns:
        _type_: _description_
    """
    # Reshape the position vectors
    surface_coordenates = {
        'wall_north': [[0,0,h],[0,0,0],[w,0,0],[w,0,h]],
        'wall_east': [[w,0,h],[w,0,0],[w,l,0],[w,l,h]],
        'wall_south': [[w,l,h],[w,l,0],[0,l,0],[0,l,h]],
        'wall_west': [[0,l,h],[0,l,0],[0,0,0],[0,0,h]],
        'roof': [[w,0,h],[w,l,h],[0,l,h],[0,0,h]],
        'floor': [[w,l,0],[w,0,0],[0,0,0],[0,l,0]]
    }
    coordinate = ['x', 'y', 'z']
    for surface_name in [key for key in surface_coordenates.keys()]:
        # Iterate over the four vertices of the surface
        for _ in range(4):
            for xyz in range(3):
                epJSON_object['BuildingSurface:Detailed'][surface_name]['vertices'][_]['vertex_'+coordinate[xyz]+'_coordinate'] = surface_coordenates[surface_name][_][xyz]
    
    north_window_proportion = window_area_relation[0]
    east_window_proportion = window_area_relation[1]
    south_window_proportion = window_area_relation[2]
    west_window_proportion = window_area_relation[3]
    
    window_coordinates = {
        'window_north': [
            [0+(w*north_window_proportion)/2,0,h*0.9],
            [0+(w*north_window_proportion)/2,0,0.1],
            [w-(w*north_window_proportion)/2,0,0.1],
            [w-(w*north_window_proportion)/2,0,h*0.9]
        ],
        'window_east': [
            [w,0+(l*east_window_proportion)/2,h*0.9],
            [w,0+(l*east_window_proportion)/2,0.1],
            [w,l-(l*east_window_proportion)/2,0.1],
            [w,l-(l*east_window_proportion)/2,h*0.9]
        ],
        'window_south': [
            [w-(w*south_window_proportion)/2,l,h*0.9],
            [w-(w*south_window_proportion)/2,l,0.1],
            [0+(w*south_window_proportion)/2,l,0.1],
            [0+(w*south_window_proportion)/2,l,h*0.9]
        ],
        'window_west': [
            [0,l-(l*west_window_proportion)/2,h*0.9],
            [0,l-(l*west_window_proportion)/2,0.1],
            [0,0+(l*west_window_proportion)/2,0.1],
            [0,0+(l*west_window_proportion)/2,h*0.9]
        ],
    }
    
    for window_name in [key for key in window_coordinates.keys()]:
        # Iterate over the four vertices of the surface
        for _ in range(1,5,1):
            for xyz in range(3):
                epJSON_object['FenestrationSurface:Detailed'][window_name]['vertex_'+str(_)+'_'+coordinate[xyz]+'_coordinate'] = window_coordinates[window_name][_-1][xyz]
        


import gymnasium as gym

class Holon:
    def __init__(self, config: dict):
        self.observation_space: gym.Space = config['observation_space']
        self.action_space: gym.Space = config['action_space']
        self._holonparts: dict = config['holonparts']
        self._superholons: dict = config['superholons']
        self._authoritylist: dict = config['authoritylist']
        
    def holonparts(self) -> dict:
        return self._holonparts
    
    def superholons(self) -> dict:
        return self._superholons
    
    def authoritylist(self) -> dict:
        return self._authoritylist
    
    def addParts(self, **kargs: dict) -> None:
        for arg in kargs:
            self._holonparts.append(arg)
    
    def removeParts(self):
        return
    def requestStructure(self):
        return
    def changeAuthorities(self):
        return
    def requestAuthorities(self):
        return
    def requestStatus(self):
        return
    def close(self):
        return